# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

import abc
from collections import namedtuple, Counter, OrderedDict

from scipy.optimize import (
    minimize, differential_evolution, basinhopping, NonlinearConstraint,
    least_squares,
)
from scipy.optimize import BFGS as soBFGS
import sympy
import numpy as np

from .fit_results import FitResults
from .objectives import BaseObjective, MinimizeModel
from .models import CallableNumericalModel, BaseModel

import inspect
from functools import wraps

DummyModel = namedtuple('DummyModel', 'params')


class BaseMinimizer(object):
    """
    ABC for all Minimizers.
    """
    def __init__(self, objective, parameters):
        """
        :param objective: Objective function to be used.
        :param parameters: List of :class:`~symfit.core.argument.Parameter` instances
        """
        self.parameters = parameters
        self._fixed_params = [p for p in parameters if p.fixed]
        self.objective = self._baseobjective_from_callable(objective)

        # Mapping which we use to track the original, to be used upon pickling
        self._pickle_kwargs = {'parameters': parameters, 'objective': objective}
        self.params = [p for p in parameters if not p.fixed]

    def _baseobjective_from_callable(self, func, objective_type=MinimizeModel):
        """
        symfit works with BaseObjective subclasses internally. If a custom
        objective is provided, we wrap it into a BaseObjective, MinimizeModel by
        default.

        :param func: Callable. If already an instance of BaseObjective, it is
            returned immediately. If not, it is turned into a BaseObjective of
            type ``objective_type``.
        :param objective_type:
        :return:
        """
        if isinstance(func, BaseObjective) or (hasattr(func, '__self__') and
                                               isinstance(func.__self__, BaseObjective)):
            # The latter condition is added to make sure .eval_jacobian methods
            # are still considered correct, and not doubly wrapped.
            return func
        else:
            if isinstance(func, BaseModel):
                model = func
            else:
                # Minimize the provided custom objective instead. We therefore
                # wrap it into a CallableNumericalModel, thats what they are for
                y = sympy.Dummy()
                model = CallableNumericalModel(
                    {y: func},
                    connectivity_mapping={y: set(self.parameters)}
                )
            return objective_type(model,
                                  data={})

    @abc.abstractmethod
    def execute(self, **options):
        """
        The execute method should implement the actual minimization procedure,
        and should return a :class:`~symfit.core.fit_results.FitResults` instance.

        :param options: options to be used by the minimization procedure.
        :return:  an instance of :class:`~symfit.core.fit_results.FitResults`.
        """
        pass

    @property
    def initial_guesses(self):
        try:
            return self._initial_guesses
        except AttributeError:
            return [p.value for p in self.params]

    @initial_guesses.setter
    def initial_guesses(self, vals):
        self._initial_guesses = vals

    def __getstate__(self):
        return {key: value for key, value in self.__dict__.items()
                if not key.startswith('wrapped_')}

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__init__(**self._pickle_kwargs)

class BoundedMinimizer(BaseMinimizer):
    """
    ABC for Minimizers that support bounds.
    """
    @property
    def bounds(self):
        return [(p.min, p.max) for p in self.params]

class ConstrainedMinimizer(BaseMinimizer):
    """
    ABC for Minimizers that support constraints
    """
    def __init__(self, *args, constraints=None, **kwargs):
        super(ConstrainedMinimizer, self).__init__(*args, **kwargs)
        # Remember the vanilla constraints for pickling
        self._pickle_kwargs['constraints'] = constraints
        if constraints is None:
            constraints = []
        self.constraints = constraints

class GradientMinimizer(BaseMinimizer):
    """
    ABC for Minizers that support the use of a jacobian
    """
    def __init__(self, *args, jacobian=None, **kwargs):
        self.jacobian = jacobian
        super(GradientMinimizer, self).__init__(*args, **kwargs)
        self._pickle_kwargs['jacobian'] = self.jacobian
        if self.jacobian is not None:
            self.jacobian = self._baseobjective_from_callable(self.jacobian)
            self.wrapped_jacobian = self.resize_jac(self.jacobian)
        else:
            self.wrapped_jacobian = None

    def resize_jac(self, func):
        """
        Removes values with identical indices to fixed parameters from the
        output of func. func has to return the jacobian of a scalar function.

        :param func: Jacobian function to be wrapped. Is assumed to be the
            jacobian of a scalar function.
        :return: Jacobian corresponding to non-fixed parameters only.
        """
        if func is None:
            return None
        @wraps(func)
        def resized(*args, **kwargs):
            out = func(*args, **kwargs)
            # Make one dimensional, corresponding to a scalar function.
            out = np.atleast_1d(np.squeeze(out))
            mask = [p not in self._fixed_params for p in self.parameters]
            return out[mask]
        return resized


class HessianMinimizer(GradientMinimizer):
    """
    ABC for Minimizers that support the use of a Hessian.
    """
    def __init__(self, *args, hessian=None, **kwargs):
        self.hessian = hessian
        super(HessianMinimizer, self).__init__(*args, **kwargs)
        self._pickle_kwargs['hessian'] = self.hessian
        if self.hessian is not None:
            self.hessian = self._baseobjective_from_callable(self.hessian)
            self.wrapped_hessian = self.resize_hess(self.hessian)
        else:
            self.wrapped_hessian = None

    def resize_hess(self, func):
        """
        Removes values with identical indices to fixed parameters from the
        output of func. func has to return the Hessian of a scalar function.

        :param func: Hessian function to be wrapped. Is assumed to be the
            Hessian of a scalar function.
        :return: Hessian corresponding to free parameters only.
        """
        if func is None:
            return None
        @wraps(func)
        def resized(*args, **kwargs):
            out = func(*args, **kwargs)
            # Make two dimensional, corresponding to a scalar function.
            out = np.atleast_2d(np.squeeze(out))
            mask = [p not in self._fixed_params for p in self.parameters]
            return np.atleast_2d(out[mask, mask])
        return resized


class GlobalMinimizer(BaseMinimizer):
    """
    A minimizer that looks for a global minimum, instead of a local one.
    """
    def __init__(self, *args, **kwargs):
        super(GlobalMinimizer, self).__init__(*args, **kwargs)


class ChainedMinimizer(BaseMinimizer):
    """
    A minimizer that consists of multiple other minimizers, each executed in
    order.
    This is valuable if you have minimizers that are not good at finding the
    exact minimum such as :class:`~symfit.core.minimizers.NelderMead` or
    :class:`~symfit.core.minimizers.DifferentialEvolution`.
    """
    def __init__(self, *args, minimizers=None, **kwargs):
        '''
        :param minimizers: a :class:`~collections.abc.Sequence` of
            :class:`~symfit.core.minimizers.BaseMinimizer` objects, which need
            to be run in order.
        :param \*args: passed to :func:`symfit.core.minimizers.BaseMinimizer.__init__`.
        :param \*\*kwargs: passed to :func:`symfit.core.minimizers.BaseMinimizer.__init__`.
        '''
        super(ChainedMinimizer, self).__init__(*args, **kwargs)
        self.minimizers = minimizers
        self._pickle_kwargs['minimizers'] = self.minimizers
        self.__signature__ = self._make_signature()

    def execute(self, **minimizer_kwargs):
        """
        Execute the chained-minimization. In order to pass options to the
        seperate minimizers, they can  be passed by using the
        names of the minimizers as keywords. For example::

            fit = Fit(self.model, self.xx, self.yy, self.ydata,
                      minimizer=[DifferentialEvolution, BFGS])
            fit_result = fit.execute(
                DifferentialEvolution={'seed': 0, 'tol': 1e-4, 'maxiter': 10},
                BFGS={'tol': 1e-4}
            )

        In case of multiple identical minimizers an index is added to each
        keyword argument to make them identifiable. For example, if::

            minimizer=[BFGS, DifferentialEvolution, BFGS])

        then the keyword arguments will be 'BFGS', 'DifferentialEvolution',
        and 'BFGS_2'.

        :param minimizer_kwargs: Minimizer options to be passed to the
            minimzers by name
        :return:  an instance of :class:`~symfit.core.fit_results.FitResults`.
        """
        bound_arguments = self.__signature__.bind(**minimizer_kwargs)
        # Include default values in bound_argument object.
        # Start from a new OrderedDict to guarantee ordering.
        arguments = OrderedDict()
        for param in self.__signature__.parameters.values():
            if param.name in bound_arguments.arguments:
                arguments[param.name] = bound_arguments.arguments[param.name]
            else:
                arguments[param.name] = param.default
        bound_arguments.arguments = arguments

        answers = []
        next_guess = self.initial_guesses
        for minimizer, kwargs in zip(self.minimizers, bound_arguments.arguments.values()):
            minimizer.initial_guesses = next_guess
            ans = minimizer.execute(**kwargs)
            next_guess = list(ans.params.values())
            answers.append(ans)
        final = answers[-1]
        # TODO: Compile all previous results in one, instead of just the
        # number of function evaluations. But there's some code down the
        # line that expects scalars.
        final.minimizer_output['nit'] = sum(ans.iterations for ans in answers)
        return final

    def _make_signature(self):
        """
        Create a signature for `execute` based on the minimizers this
        `ChainedMinimizer` was initiated with. For the format, see the docstring
        of :meth:`ChainedMinimizer.execute`.

        :return: :class:`inspect.Signature` instance.
        """
        # Create KEYWORD_ONLY arguments with the names of the minimizers.
        name = lambda x: x.__class__.__name__
        count = Counter(
            [name(minimizer) for minimizer in self.minimizers]
        ) # Count the number of each minimizer, they don't have to be unique

        # Note that these are inspect.Parameter's, not symfit parameters!
        parameters = []
        for minimizer in reversed(self.minimizers):
            if count[name(minimizer)] == 1:
                # No ambiguity, so use the name directly.
                param_name = name(minimizer)
            else:
                # Ambiguity, so append the number of remaining minimizers
                param_name = '{}_{}'.format(name(minimizer), count[name(minimizer)])
            count[name(minimizer)] -= 1

            parameters.append(
                inspect.Parameter(
                    param_name,
                    kind=inspect.Parameter.KEYWORD_ONLY,
                    default={}
                )
            )
        return inspect.Signature(parameters=reversed(parameters))

    def __getstate__(self):
        state = super(ChainedMinimizer, self).__getstate__()
        del state['__signature__']
        return state

    def __str__(self):
        return self.__class__.__name__ + '(minimizers={})'.format(self.minimizers)

class ScipyMinimize(object):
    """
    Mix-in class that handles the execute calls to :func:`scipy.optimize.minimize`.
    """
    def __init__(self, *args, **kwargs):
        self.constraints = []
        self.jacobian = None
        self.wrapped_jacobian = None
        super(ScipyMinimize, self).__init__(*args, **kwargs)

    def execute(self, bounds=None, jacobian=None, hessian=None,
                constraints=None, *, tol=1e-9, **minimize_options):
        """
        Calls the wrapped algorithm.

        :param bounds: The bounds for the parameters. Usually filled by
            :class:`~symfit.core.minimizers.BoundedMinimizer`.
        :param jacobian: The Jacobian. Usually filled by
            :class:`~symfit.core.minimizers.ScipyGradientMinimize`.
        :param \*\*minimize_options: Further keywords to pass to
            :func:`scipy.optimize.minimize`. Note that your `method` will
            usually be filled by a specific subclass.
        """
        minimize_options.update(tol=tol)
        ans = minimize(
            self.objective,
            self.initial_guesses,
            method=self.method_name(),
            bounds=bounds,
            constraints=constraints,
            jac=jacobian,
            hess=hessian,
            **minimize_options
        )
        return self._pack_output(ans)

    def _pack_output(self, ans):
        """
        Packs the output of a minimization in a
        :class:`~symfit.core.fit_results.FitResults`.

        :param ans: The output of a minimization as produced by
            :func:`scipy.optimize.minimize`
        :returns: :class:`~symfit.core.fit_results.FitResults`
        """
        best_vals = []
        found = iter(np.atleast_1d(ans.x))
        for param in self.parameters:
            if param.fixed:
                best_vals.append(param.value)
            else:
                best_vals.append(next(found))

        fit_results = dict(
            model=DummyModel(params=self.parameters),
            popt=best_vals,
            covariance_matrix=None,
            objective=self.objective,
            minimizer=self,
            **ans
        )

        return FitResults(**fit_results)

    @classmethod
    def method_name(cls):
        """
        Returns the name of the minimize method this object represents. This is
        needed because the name of the object is not always exactly what needs
        to be passed on to scipy as a string.
        :return:
        """
        return cls.__name__


class ScipyGradientMinimize(ScipyMinimize, GradientMinimizer):
    """
    Base class for :func:`scipy.optimize.minimize`'s gradient-minimizers.
    """
    def execute(self, *, jacobian=None, **minimize_options):
        # This method takes the jacobian as an argument because the user may
        # need to override it in some cases (especially with the trust-constr 
        # method)
        if jacobian is None:
            jacobian = self.wrapped_jacobian
        return super(ScipyGradientMinimize, self).execute(jacobian=jacobian, **minimize_options)

    def scipy_constraints(self, constraints):
        cons = super(ScipyGradientMinimize, self).scipy_constraints(constraints)
        for con in cons:
            # Only if the model has a jacobian, does it make sense to pass one
            # to the minimizer
            if hasattr(con['fun'].model, 'eval_jacobian'):
                con['jac'] = self.resize_jac(con['fun'].eval_jacobian)
            else:
                con['jac'] = None
        return cons


class ScipyBoundedMinimizer(ScipyMinimize, BoundedMinimizer):
    """
    Base class for :func:`scipy.optimize.minimize`'s bounded-minimizers.
    """
    def execute(self, **minimize_options):
        return super(ScipyBoundedMinimizer, self).execute(bounds=self.bounds,
                                                          **minimize_options)


class ScipyHessianMinimize(ScipyGradientMinimize, HessianMinimizer):
    """
    Base class for :func:`scipy.optimize.minimize`'s hessian-minimizers.
    """
    def execute(self, *, hessian=None, **minimize_options):
        # This method takes the hessian as an argument because the user may
        # need to override it in some cases (especially with the trust-constr 
        # method)
        if hessian is None:
            hessian = self.wrapped_hessian
        return super(ScipyHessianMinimize, self).execute(hessian=hessian, **minimize_options)

    def scipy_constraints(self, constraints):
        cons = super(ScipyHessianMinimize, self).scipy_constraints(constraints)
        for con in cons:
            # Only if the model has a hessian, does it make sense to pass one
            # to the minimizer
            if hasattr(con['fun'].model, 'eval_hessian'):
                con['hess'] = self.resize_hess(con['fun'].eval_hessian)
            else:
                con['hess'] = None
        return cons


class ScipyConstrainedMinimize(ScipyMinimize, ConstrainedMinimizer):
    """
    Base class for :func:`scipy.optimize.minimize`'s constrained-minimizers.
    """
    def __init__(self, *args, **kwargs):
        super(ScipyConstrainedMinimize, self).__init__(*args, **kwargs)
        self.wrapped_constraints = self.scipy_constraints(self.constraints)

    def execute(self, **minimize_options):
        return super(ScipyConstrainedMinimize, self).execute(constraints=self.wrapped_constraints, **minimize_options)

    def scipy_constraints(self, constraints):
        """
        Returns all constraints in a scipy compatible format.

        :param constraints: List of either MinimizeModel instances (this is what
          is provided by :class:`~symfit.core.fit.Fit`),
          :class:`~symfit.core.models.BaseModel`, or
          :class:`sympy.core.relational.Relational`.
        :return: dict of scipy compatible statements.
        """
        cons = []
        types = {  # scipy only distinguishes two types of constraint.
            sympy.Eq: 'eq', sympy.Ge: 'ineq',
        }

        for constraint in constraints:
            if isinstance(constraint, MinimizeModel):
                # Typically the case when called by `Fit
                constraint_type = constraint.model.constraint_type
            elif hasattr(constraint, 'constraint_type'):
                # Model object, not provided by `Fit`. Do the best we can.
                if self.parameters != constraint.params:
                    raise AssertionError('The constraint should accept the same'
                                         ' parameters as used for the fit.')
                constraint_type = constraint.constraint_type
                constraint = MinimizeModel(constraint, data=self.objective.data)
            elif isinstance(constraint, sympy.Rel):
                constraint_type = constraint.__class__
                constraint = self.objective.model.__class__.as_constraint(
                    constraint, self.objective.model
                )
                constraint = MinimizeModel(constraint, data=self.objective.data)
            else:
                raise TypeError('Unknown type for a constraint.')
            con = {
                'type': types[constraint_type],
                'fun': constraint,
                }
            cons.append(con)
        cons = tuple(cons)
        return cons

    def _pack_output(self, ans):
        fit_result = super(ScipyConstrainedMinimize, self)._pack_output(ans)
        fit_result.constraints = self.constraints
        return fit_result


class BFGS(ScipyGradientMinimize):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s BFGS algorithm.
    """


class SLSQP(ScipyGradientMinimize, ScipyConstrainedMinimize, ScipyBoundedMinimizer):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s SLSQP algorithm.
    """


class COBYLA(ScipyConstrainedMinimize, BaseMinimizer):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s COBYLA algorithm.
    """
    def execute(self, **minimize_options):
        ans = super(COBYLA, self).execute(**minimize_options)
        # Nearest indication of nit.
        ans.minimizer_output['nit'] = ans.minimizer_output.pop('nfev')
        return ans

class LBFGSB(ScipyGradientMinimize, ScipyBoundedMinimizer):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s LBFGSB algorithm.
    """
    @classmethod
    def method_name(cls):
        return "L-BFGS-B"

class NelderMead(ScipyMinimize, BaseMinimizer):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s NelderMead algorithm.
    """
    @classmethod
    def method_name(cls):
        return 'Nelder-Mead'

class Powell(ScipyMinimize, BaseMinimizer):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s Powell algorithm.
    """

class TrustConstr(ScipyHessianMinimize, ScipyConstrainedMinimize, ScipyBoundedMinimizer):
    """
    Wrapper around :func:`scipy.optimize.minimize`'s Trust-Constr algorithm.
    """
    @classmethod
    def method_name(cls):
        return 'trust-constr'

    def _get_jacobian_hessian_strategy(self):
        """
        Figure out how to calculate the jacobian and hessian. Will return a
        tuple describing how best to calculate the jacobian and hessian,
        repectively. If None, it should be calculated using the available
        analytical method.

        :return: tuple of jacobian_method, hessian_method
        """
        if self.jacobian is not None and self.hessian is None:
            jacobian = None
            hessian = 'cs'
        elif self.jacobian is None and self.hessian is None:
            jacobian = 'cs'
            hessian = soBFGS(exception_strategy='damp_update')
        else:
            jacobian = None
            hessian = None
        return jacobian, hessian

    def scipy_constraints(self, constraints):
        cons = super(TrustConstr, self).scipy_constraints(constraints)
        out = []
        for con in cons:
            if con['type'] == 'eq':
                ub = 0
            else:
                ub = np.inf

            nonlinearconstr_kwargs = {
                'fun': con['fun'], 'lb': 0, 'ub': ub,
            }
            # If None is given to NonlinearConstraint it'll throw a hissy fit.
            if con['hess'] is not None:
                nonlinearconstr_kwargs['hess'] = lambda x, v: con['hess'](x) * v
            if con['jac'] is not None:
                nonlinearconstr_kwargs['jac'] = con['jac']
            tc_con = NonlinearConstraint(**nonlinearconstr_kwargs)
            out.append(tc_con)
        return out

    def execute(self, *, jacobian=None, hessian=None, options=None, **minimize_options):
        if options is None:
            options = {}
        # Our Jacobians are dense, and apparently we need to explicitely
        # tell this.
        options['sparse_jacobian'] = False

        auto_jacobian, auto_hessian = self._get_jacobian_hessian_strategy()
        # For models that are not differentiable, users need the ability to
        # change the jacobian to e.g. 'cs' or '3-point'. In that case, hess
        # should either be scipy.optimize.BFGS or SR1.
        # In addition, users may want to change the way the Hessian is
        # calculated, especially if they manage to make a model whose Jacobian
        # can't handle complex numbers.
        if jacobian is None:
            jacobian = auto_jacobian
        if hessian is None:
            hessian = auto_hessian

        if jacobian is None:
            jacobian = self.wrapped_jacobian
        if hessian is None:
            hessian = self.wrapped_hessian

        ans = super(TrustConstr, self).execute(options=options,
                                                jacobian=jacobian,
                                                hessian=hessian,
                                                **minimize_options)
        # Rename the number of iterations kwarg to be consistent.
        ans.minimizer_output['nit'] = ans.minimizer_output.pop('niter')
        return ans


class DifferentialEvolution(ScipyBoundedMinimizer, GlobalMinimizer):
    """
    A wrapper around :func:`scipy.optimize.differential_evolution`.
    """
    def execute(self, *, strategy='rand1bin', popsize=40, mutation=(0.423, 1.053),
                recombination=0.95, polish=False, init='latinhypercube', **de_options):
        de_options.update(
            strategy=strategy, popsize=popsize, mutation=mutation,
            recombination=recombination, polish=polish, init=init
        )
        ans = differential_evolution(self.objective,
                                     self.bounds,
                                     **de_options)
        return self._pack_output(ans)


class BasinHopping(ScipyMinimize, GlobalMinimizer):
    """
    Wrapper around :func:`scipy.optimize.basinhopping`'s basin-hopping algorithm.

    As always, the best way to use this algorithm is through
    :class:`~symfit.core.fit.Fit`, as this will automatically select a local
    minimizer for you depending on whether you provided bounds, constraints, etc.

    However, BasinHopping can also be used directly. Example (with jacobian)::

        import numpy as np
        from symfit.core.minimizers import BFGS, BasinHopping
        from symfit import parameters

        def func2d(x1, x2):
            f = np.cos(14.5 * x1 - 0.3) + (x2 + 0.2) * x2 + (x1 + 0.2) * x1
            return f

        def jac2d(x1, x2):
            df = np.zeros(2)
            df[0] = -14.5 * np.sin(14.5 * x1 - 0.3) + 2. * x1 + 0.2
            df[1] = 2. * x2 + 0.2
            return df

        x0 = [1.0, 1.0]
        np.random.seed(555)
        x1, x2 = parameters('x1, x2', value=x0)
        fit = BasinHopping(func2d, [x1, x2], local_minimizer=BFGS)
        minimizer_kwargs = {'jac': fit.list2kwargs(jac2d)}
        fit_result = fit.execute(niter=200, minimizer_kwargs=minimizer_kwargs)

    See :func:`scipy.optimize.basinhopping` for more options.
    """
    def __init__(self, *args, local_minimizer=BFGS, **kwargs):
        """
        :param local_minimizer: minimizer to be used for local minimization
            steps. Can be any subclass of
            :class:`symfit.core.minimizers.ScipyMinimize`.
        :param args: positional arguments to be passed on to `super`.
        :param kwargs: keyword arguments to be passed on to `super`.
        """
        self.local_minimizer = local_minimizer
        super(BasinHopping, self).__init__(*args, **kwargs)
        self._pickle_kwargs['local_minimizer'] = self.local_minimizer

        type_error_msg = 'Currently only subclasses of ScipyMinimize are ' \
                         'supported, since `scipy.optimize.basinhopping` uses ' \
                         '`scipy.optimize.minimize`.'
        # self.local_minimizer has to be a subclass or instance of ScipyMinimize
        # Since no one function exists to test this, we try/except instead.
        try:
            # Test if subclass. If this line doesn't fail, we are dealing with
            # some class. If it fails, we assume that it is an instance.
            issubclass(self.local_minimizer, ScipyMinimize)
        except TypeError:
            # It is not a class at all, so test if it's an instance instead
            if not isinstance(self.local_minimizer, ScipyMinimize):
                # Only ScipyMinimize subclasses supported
                raise TypeError(type_error_msg)
        else:
            if not issubclass(self.local_minimizer, ScipyMinimize):
                # Only ScipyMinimize subclasses supported
                raise TypeError(type_error_msg)
            self.local_minimizer = self.local_minimizer(self.objective, self.parameters)

    def execute(self, **minimize_options):
        """
        Execute the basin-hopping minimization.

        :param minimize_options: options to be passed on to
            :func:`scipy.optimize.basinhopping`.
        :return: :class:`symfit.core.fit_results.FitResults`
        """
        if 'minimizer_kwargs' not in minimize_options:
            minimize_options['minimizer_kwargs'] = {}

        if 'method' not in minimize_options['minimizer_kwargs']:
            # If no minimizer was set by the user upon execute, use local_minimizer
            minimize_options['minimizer_kwargs']['method'] = self.local_minimizer.method_name()
        if 'jac' not in minimize_options['minimizer_kwargs'] and isinstance(self.local_minimizer, GradientMinimizer):
            # Assign the jacobian
            minimize_options['minimizer_kwargs']['jac'] = self.local_minimizer.wrapped_jacobian
        if 'constraints' not in minimize_options['minimizer_kwargs'] and isinstance(self.local_minimizer, ConstrainedMinimizer):
            # Assign constraints
            minimize_options['minimizer_kwargs']['constraints'] = self.local_minimizer.wrapped_constraints
        if 'bounds' not in minimize_options['minimizer_kwargs'] and isinstance(self.local_minimizer, BoundedMinimizer):
            # Assign bounds
            minimize_options['minimizer_kwargs']['bounds'] = self.local_minimizer.bounds

        ans = basinhopping(
            self.objective,
            self.initial_guesses,
            **minimize_options
        )
        if isinstance(ans.message, list):
            # For some reason this is currently a length one list containing
            # the message. We check just in case this gets fixed upstream in
            # future releases.
            ans.message = ans.message[0]
        if 'constraints' in minimize_options['minimizer_kwargs']:
            # Add the constraints to the FitResults
            ans['constraints'] = self.local_minimizer.constraints
        return self._pack_output(ans)


class MINPACK(ScipyBoundedMinimizer, GradientMinimizer):
    """
    Wrapper to scipy's implementation of least_squares, since it is the industry
    standard.
    """
    def resize_jac(self, func):
        """
        Removes values with identical indices to fixed parameters from the
        output of func. func has to return the jacobian of the residuals.
        This method is different from the one in GradientMinimizer, since
        least_squares expects the jacobian to return an MxN (M=len(data),
        N=len(params)) matrix, rather than a vector.

        :param func: Jacobian function to be wrapped. Is assumed to be the
            jacobian of the residuals.
        :return: Jacobian corresponding to non-fixed parameters only.
        """
        if func is None:
            return None
        @wraps(func)
        def resized(*args, **kwargs):
            out = func(*args, **kwargs)
            out = np.atleast_2d(out)
            mask = [p not in self._fixed_params for p in self.parameters]
            return out[:, mask]
        return resized

    @property
    def bounds(self):
        lbounds = []
        ubounds = []
        for low, high in super().bounds:
            if low is None:
                low = -np.inf
            if high is None:
                high = np.inf
            lbounds.append(low)
            ubounds.append(high)
        return lbounds, ubounds

    def execute(self, jacobian=None, method='trf', **minpack_options):
        """
        :param \*\*minpack_options: Any named arguments to be passed to
            :func:`scipy.optimize.least_squares`
        """
        if jacobian is None:
            jacobian = self.wrapped_jacobian
        jacobian = jacobian or 'cs'

        if not self.bounds:
            method = method or 'lm'
        else:
            method = method or 'trf'

        full_output = least_squares(
            self.objective,
            x0=self.initial_guesses,
            jac=jacobian,
            bounds=self.bounds,
            method=method,
            **minpack_options
        )
        return self._pack_output(full_output)



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

# -*- coding: utf-8 -*-

from ... import ODEModel, Derivative, latex
from ...core.fit import TakesData
from ...core.support import key2str, deprecated

import itertools

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import gaussian_kde

plt.ioff()


class InteractiveGuess(TakesData):
    """A class that provides an graphical, interactive way of guessing initial
    fitting parameters."""

    def __init__(self, *args, n_points=50, log_contour=True, percentile=(5, 95), **kwargs):
        """Create a matplotlib window with sliders for all parameters
        in this model, so that you may graphically guess initial fitting
        parameters. n_points is the number of points drawn for the plot.
        Data points are plotted as a blue contour plot, the proposed model as
        a red line. The errorbars on the proposed model represent the
        percentile of data within the thresholds.

        Slider extremes are taken from the parameters where possible. If
        these are not provided, the minimum is 0; and the maximum is value*2.
        If no initial value is provided, it defaults to 1.

        This will modify the values of the parameters present in model.

        :param n_points: The number of points used for drawing the
            fitted function. Defaults to 50.
        :type n_points: int

        :param log_contour: Whether to plot the contour plot of the log of the
            density, rather than the density itself. If True, any density less
            than 1e-7 will be considered 0. Defaults to True.
        :type log_contour: bool

        :param percentile: Controls the errorbars on the proposed model, such
            that the lower errorbar will cover percentile[0]% of the data, and
            the upper will cover percentile[1]%. Defaults to [5, 95], with
            corresponds to a 90% percentile. Should be a list of 2 numbers.
        :type percentile: list
        """
        self.log_contour = log_contour
        self.percentile = percentile
        super(InteractiveGuess, self).__init__(*args, **kwargs)
        if len(self.independent_data) > 1:
            self._dimension_strategy = StrategynD(self)
        else:
            self._dimension_strategy = Strategy2D(self)
        # TODO: Some of the code here is specific to the n-D case and should
        # be moved.
        self._projections = list(itertools.product(self.model.independent_vars,
                                                   self.model.dependent_vars))
        x_mins = {v: np.min(data) for v, data in self.independent_data.items()}
        x_maxs = {v: np.max(data) for v, data in self.independent_data.items()}

        # Stretch the plot 10-20% in the X direction, since that is visually
        # more appealing. We can't evaluate the model for x < x_initial, so
        # don't.
        for x in self.model.independent_vars:
            plotrange_x = x_maxs[x] - x_mins[x]
            if not hasattr(self.model, 'initial'):
                x_mins[x] -= 0.1 * plotrange_x
            x_maxs[x] += 0.1 * plotrange_x
        # Generate the points at which to evaluate the model with the proposed
        # parameters for plotting
        self._x_points = {v: np.linspace(x_mins[v], x_maxs[v], n_points)
                          for v in self.independent_data}
        meshgrid = np.meshgrid(*(self._x_points[v]
                                 for v in self.independent_data))
        self._x_grid = {v: meshgrid[idx].flatten()
                        for idx, v in enumerate(self.independent_data)}

        # Stretch the plot 20% in the Y direction, since that is visually more
        # appealing
        y_mins = {v: np.min(data) for v, data in self.dependent_data.items()}
        y_maxs = {v: np.max(data) for v, data in self.dependent_data.items()}
        for y in self.dependent_data:
            plotrange_y = y_maxs[y] - y_mins[y]
            y_mins[y] -= 0.1 * plotrange_y
            y_maxs[y] += 0.1 * plotrange_y

        self._y_points = {v: np.linspace(y_mins[v], y_maxs[v], n_points)
                          for v in self.dependent_data}

        self._set_up_figure(x_mins, x_maxs, y_mins, y_maxs)
        self._set_up_sliders()
        self._update_plot(None)

    def execute(self, *, show=True, block=True, **kwargs):
        """
        Execute the interactive guessing procedure.

        :param show: Whether or not to show the figure. Useful for testing.
        :type show: bool
        :param block: Blocking call to matplotlib
        :type show: bool

        Any additional keyword arguments are passed to
        matplotlib.pyplot.show().
        """
        kwargs.update(block=block)
        if show:
            # self.fig.show()  # Apparently this does something else,
            # see https://github.com/matplotlib/matplotlib/issues/6138
            plt.show(**kwargs)

    def _set_up_figure(self, x_mins, x_maxs, y_mins, y_maxs):
        """
        Prepare the matplotlib figure: make all the subplots; adjust their
        x and y range; plot the data; and plot an putative function.
        """
        self.fig = plt.figure()

        # Make room for the sliders:
        bot = 0.1 + 0.05*len(self.model.params)
        self.fig.subplots_adjust(bottom=bot)

        # If these are not ints, matplotlib will crash and burn with an utterly
        # vague error.
        nrows = int(np.ceil(len(self._projections)**0.5))
        ncols = int(np.ceil(len(self._projections)/nrows))

        # Make all the subplots: set the x and y limits, scatter the data, and
        # plot the putative function.
        self._plots = {}

        for plotnr, proj in enumerate(self._projections, 1):
            x, y = proj
            if Derivative(y, x) in self.model:
                title_format = '$\\frac{{\\partial {dependant}}}{{\\partial {independant}}} = {expression}$'
            else:
                title_format = '${dependant}({independant}) = {expression}$'
            plotlabel = title_format.format(
                dependant=latex(y, mode='plain'),
                independant=latex(x, mode='plain'),
                expression=latex(self.model[y], mode='plain'))
            ax = self.fig.add_subplot(ncols, nrows, plotnr,
                                      label=plotlabel)
            ax.set_title(ax.get_label())
            ax.set_ylim(y_mins[y], y_maxs[y])
            ax.set_xlim(x_mins[x], x_maxs[x])
            ax.set_xlabel('${}$'.format(x))
            ax.set_ylabel('${}$'.format(y))
            self._plot_data(proj, ax)
            plot = self._plot_model(proj, ax)
            self._plots[proj] = plot

    def _set_up_sliders(self):
        """
        Creates an slider for every parameter.
        """
        i = 0.05
        self._sliders = {}
        for param in self.model.params:
            if not param.fixed:
                axbg = 'lightgoldenrodyellow'
            else:
                axbg = 'red'
            # start-x, start-y, width, height
            ax = self.fig.add_axes((0.162, i, 0.68, 0.03),
                                   facecolor=axbg, label=param)
            val = param.value
            if not hasattr(param, 'min') or param.min is None:
                minimum = 0
            else:
                minimum = param.min
            if not hasattr(param, 'max') or param.max is None:
                maximum = 2 * val
            else:
                maximum = param.max

            slid = plt.Slider(ax, param, minimum, maximum,
                              valinit=val, valfmt='% 5.4g')
            self._sliders[param] = slid
            slid.on_changed(self._update_plot)
            i += 0.05

    def _plot_data(self, proj, ax):
        """Defers plotting the data to self._dimension_strategy"""
        return self._dimension_strategy.plot_data(proj, ax)

    def _plot_model(self, proj, ax):
        """Defers plotting the proposed model to self._dimension_strategy"""
        return self._dimension_strategy.plot_model(proj, ax)

    def _update_specific_plot(self, indep_var, dep_var):
        """Defers updating the proposed model to self._dimension_strategy"""
        return self._dimension_strategy.update_plot(indep_var, dep_var)

    def _update_plot(self, _):
        """Callback to redraw the plot to reflect the new parameter values."""
        # Since all sliders call this same callback without saying who they are
        # I need to update the values for all parameters. This can be
        # circumvented by creating a seperate callback function for each
        # parameter.
        for param in self.model.params:
            param.value = self._sliders[param].val
        for indep_var, dep_var in self._projections:
            self._update_specific_plot(indep_var, dep_var)

    def _eval_model(self):
        """
        Convenience method for evaluating the model with the current parameters

        :return: named tuple with results
        """
        arguments = self._x_grid.copy()
        arguments.update({param: param.value for param in self.model.params})
        return self.model(**key2str(arguments))

    def __str__(self):
        """
        Represent the guesses in a human readable way.

        :return: string with the guessed values.
        """
        msg = 'Guessed values:\n'
        for param in self.model.params:
            msg += '{}: {}\n'.format(param.name, param.value)
        return msg


class Strategy2D:
    """
    A strategy that describes how to plot a model that depends on a single independent variable,
    and how to update that plot.
    """
    def __init__(self, interactive_guess):
        self.ig = interactive_guess

    def plot_data(self, proj, ax):
        """
        Creates and plots a scatter plot of the original data.
        """
        x, y = proj
        ax.scatter(self.ig.independent_data[x],
                   self.ig.dependent_data[y], c='b')

    def plot_model(self, proj, ax):
        """
        Plots the model proposed for the projection proj on ax.
        """
        x, y = proj
        y_vals = getattr(self.ig._eval_model(), y.name)
        x_vals = self.ig._x_points[x]
        plot, = ax.plot(x_vals, y_vals, c='red')
        return plot

    def update_plot(self, indep_var, dep_var):
        """
        Updates the plot of the proposed model.
        """
        evaluated_model = self.ig._eval_model()
        plot = self.ig._plots[(indep_var, dep_var)]
        y_vals = getattr(evaluated_model, dep_var.name)
        x_vals = self.ig._x_points[indep_var]
        plot.set_data(x_vals, y_vals)


class StrategynD:
    """
    A strategy that describes how to plot a model that depends on a multiple independent variables,
    and how to update that plot.
    """
    def __init__(self, interactive_guess):
        self.ig = interactive_guess

    def plot_data(self, proj, ax):
        """
        Creates and plots the contourplot of the original data. This is done
        by evaluating the density of projected datapoints on a grid.
        """
        x, y = proj
        x_data = self.ig.independent_data[x]
        y_data = self.ig.dependent_data[y]
        projected_data = np.column_stack((x_data, y_data)).T
        kde = gaussian_kde(projected_data)

        xx, yy = np.meshgrid(self.ig._x_points[x], self.ig._y_points[y])
        x_grid = xx.flatten()
        y_grid = yy.flatten()

        contour_grid = kde.pdf(np.column_stack((x_grid, y_grid)).T)
        # This is an fugly kludge, but it seems nescessary to make low density
        # areas show up.
        if self.ig.log_contour:
            contour_grid = np.log(contour_grid)
            vmin = -7
        else:
            vmin = None
        ax.contourf(xx, yy, contour_grid.reshape(xx.shape),
                    50, vmin=vmin, cmap='Blues')

    def plot_model(self, proj, ax):
        """
        Plots the model proposed for the projection proj on ax.
        """
        x, y = proj
        evaluated_model = self.ig._eval_model()
        y_vals = getattr(evaluated_model, y.name)
        x_vals = self.ig._x_grid[x]
        plot = ax.errorbar(x_vals, y_vals, xerr=0, yerr=0, c='red')
        return plot

    def update_plot(self, indep_var, dep_var):
        """
        Updates the plot of the proposed model.
        """
        evaluated_model = self.ig._eval_model()
        y_vals = getattr(evaluated_model, dep_var.name)
        x_vals = self.ig._x_grid[indep_var]

        x_plot_data = []
        y_plot_data = []
        y_plot_error = []
        # TODO: Numpy magic
        # We need the error interval for every plotted point, so find all
        # the points plotted at x=x_i, and do some statistics on those.
        # Since all the points are on a grid made by meshgrid, the error
        # in x will alwys be 0.
        for x_val in self.ig._x_points[indep_var]:
            # We get away with this instead of digitize because x_vals is
            # on a grid made with meshgrid
            idx_mask = x_vals == x_val
            xs = x_vals[idx_mask]
            x_plot_data.append(xs[0])
            ys = y_vals[idx_mask]
            y_plot_data.append(np.mean(ys))
            y_error = np.percentile(ys, self.ig.percentile)
            y_plot_error.append(y_error)

        x_plot_data = np.array(x_plot_data)
        y_plot_data = np.array(y_plot_data)
        y_plot_error = np.array(y_plot_error)

        xs = np.column_stack((x_plot_data, x_plot_data))
        yerr = y_plot_error + y_plot_data[:, np.newaxis]
        y_segments = np.dstack((xs, yerr))
        plot_line, caps, error_lines = self.ig._plots[(indep_var, dep_var)]
        plot_line.set_data(x_plot_data, y_plot_data)
        error_lines[1].set_segments(y_segments)


class InteractiveGuess2D(InteractiveGuess):
    @deprecated(InteractiveGuess)
    def __init__(self, *args, **kwargs):
        # Deprecated as of 01/06/2017
        super(InteractiveGuess2D, self).__init__(*args, **kwargs)



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

from __future__ import division, print_function
import pytest
import numpy as np
from symfit import (
    variables, parameters, Fit, Parameter, Variable,
    Equality, Model, GradientModel
)
from symfit.core.minimizers import BFGS, MINPACK, SLSQP, LBFGSB
from symfit.distributions import Gaussian


def test_vector_fitting():
    """
    Test the behavior in the presence of bounds or constraints: `Fit` should
    select `ConstrainedNumericalLeastSquares` when bounds or constraints are
    provided, or for vector models in general. For scalar models, use
    `NumericalLeastSquares`.
    """
    a, b = parameters('a, b')
    a_i, = variables('a_i')

    xdata = np.array([
        [10.1, 9., 10.5, 11.2, 9.5, 9.6, 10.],
        [102.1, 101., 100.4, 100.8, 99.2, 100., 100.8],
        [71.6, 73.2, 69.5, 70.2, 70.8, 70.6, 70.1],
    ])

    # Make a new scalar model.
    scalar_model = {a_i: a + b}
    simple_fit = Fit(
        model=scalar_model,
        a_i=xdata[0],
        minimizer=MINPACK
    )
    assert isinstance(simple_fit.minimizer, MINPACK)

    constrained_fit = Fit(
        model=scalar_model,
        a_i=xdata[0],
        constraints=[Equality(a + b, 110)]
    )
    assert isinstance(constrained_fit.minimizer, SLSQP)

    a.min = 0
    a.max = 25
    a.value = 10
    b.min = 80
    b.max = 120
    b.value = 100
    bound_fit = Fit(
        model=scalar_model,
        a_i=xdata[0],
    )
    assert isinstance(bound_fit.minimizer, LBFGSB)

    # Repeat all of the above for the Vector model
    a, b, c = parameters('a, b, c')
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = {a_i: a, b_i: b, c_i: c}

    simple_fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
    )
    assert isinstance(simple_fit.minimizer, BFGS)

    constrained_fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
        constraints=[Equality(a + b + c, 180)]
    )
    assert isinstance(constrained_fit.minimizer, SLSQP)

    a.min = 0
    a.max = 25
    a.value = 10
    b.min = 80
    b.max = 120
    b.value = 100
    bound_fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
    )
    assert isinstance(bound_fit.minimizer, LBFGSB)

    fit_result = bound_fit.execute()
    assert fit_result.value(a) == pytest.approx(np.mean(xdata[0]), rel=1e-6)
    assert fit_result.value(b) == pytest.approx(np.mean(xdata[1]), rel=1e-6)
    assert fit_result.value(c) == pytest.approx(np.mean(xdata[2]), rel=1e-6)


def test_vector_fitting_bounds():
    """
    Tests fitting to a 3 component vector valued function, with bounds.
    """
    a, b, c = parameters('a, b, c')
    a.min = 0
    a.max = 25
    b.min = 0
    b.max = 500
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = {a_i: a, b_i: b, c_i: c}

    xdata = np.array([
        [10.1, 9., 10.5, 11.2, 9.5, 9.6, 10.],
        [102.1, 101., 100.4, 100.8, 99.2, 100., 100.8],
        [71.6, 73.2, 69.5, 70.2, 70.8, 70.6, 70.1],
    ])

    fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
    )
    fit_result = fit.execute()

    assert fit_result.value(a) == pytest.approx(np.mean(xdata[0]), rel=1e-4)
    assert fit_result.value(b) == pytest.approx(np.mean(xdata[1]), rel=1e-4)
    assert fit_result.value(c) == pytest.approx(np.mean(xdata[2]), rel=1e-4)


def test_vector_fitting_guess():
    """
    Tests fitting to a 3 component vector valued function, with guesses.
    """
    a, b, c = parameters('a, b, c')
    a.value = 10
    b.value = 100
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = {a_i: a, b_i: b, c_i: c}

    xdata = np.array([
        [10.1, 9., 10.5, 11.2, 9.5, 9.6, 10.],
        [102.1, 101., 100.4, 100.8, 99.2, 100., 100.8],
        [71.6, 73.2, 69.5, 70.2, 70.8, 70.6, 70.1],
    ])

    fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
    )
    fit_result = fit.execute()

    assert fit_result.value(a) == pytest.approx(np.mean(xdata[0]), rel=1e-4)
    assert fit_result.value(b) == pytest.approx(np.mean(xdata[1]), rel=1e-4)
    assert fit_result.value(c) == pytest.approx(np.mean(xdata[2]), rel=1e-4)


def test_global_fitting():
    """
    In case of shared parameters between the components of the model, `Fit`
    should automatically use `ConstrainedLeastSquares`.
    :return:
    """
    x_1, x_2, y_1, y_2 = variables('x_1, x_2, y_1, y_2')
    y0, a_1, a_2, b_1, b_2 = parameters('y0, a_1, a_2, b_1, b_2')

    # The following vector valued function links all the equations together
    # as stated in the intro.
    model = Model({
        y_1: a_1 * x_1**2 + b_1 * x_1 + y0,
        y_2: a_2 * x_2**2 + b_2 * x_2 + y0,
    })
    assert model.shared_parameters

    # Generate data from this model
    xdata1 = np.linspace(0, 10)
    xdata2 = xdata1[::2]  # Only every other point.

    ydata1, ydata2 = model(x_1=xdata1, x_2=xdata2, a_1=101.3,
                           b_1=0.5, a_2=56.3, b_2=1.1111, y0=10.8)
    # Add some noise to make it appear like real data
    np.random.seed(1)
    ydata1 += np.random.normal(0, 2, size=ydata1.shape)
    ydata2 += np.random.normal(0, 2, size=ydata2.shape)

    xdata = [xdata1, xdata2]
    ydata = [ydata1, ydata2]

    # Guesses
    a_1.value = 100
    a_2.value = 50
    b_1.value = 1
    b_2.value = 1
    y0.value = 10

    fit = Fit(
        model, x_1=xdata[0], x_2=xdata[1], y_1=ydata[0], y_2=ydata[1]
    )
    assert isinstance(fit.minimizer, BFGS)

    # The next model does not share parameters, but is still a vector
    model = Model({
        y_1: a_1 * x_1**2 + b_1 * x_1,
        y_2: a_2 * x_2**2 + b_2 * x_2,
    })
    fit = Fit(
        model, x_1=xdata[0], x_2=xdata[1], y_1=ydata[0], y_2=ydata[1]
    )
    assert not model.shared_parameters
    assert isinstance(fit.minimizer, BFGS)

    # Scalar model, still use bfgs.
    model = Model({
        y_1: a_1 * x_1**2 + b_1 * x_1,
    })
    fit = Fit(model, x_1=xdata[0], y_1=ydata[0])
    assert model.shared_parameters is False
    assert isinstance(fit.minimizer, BFGS)


def test_gaussian_2d_fitting():
    """
    Tests fitting to a scalar gaussian function with 2 independent
    variables.
    """
    mean = (0.6, 0.4)  # x, y mean 0.6, 0.4
    cov = [[0.2**2, 0], [0, 0.1**2]]

    # TODO: Since we bin this data later on in a 100 bins, just evaluate 100
    #       points on a Gaussian, and add an appropriate amount of noise. This
    #       burns CPU cycles without good reason.
    data = np.random.multivariate_normal(mean, cov, 1000000)

    # Insert them as y,x here as np fucks up cartesian conventions.
    ydata, xedges, yedges = np.histogram2d(data[:, 0], data[:, 1], bins=100,
                                           range=[[0.0, 1.0], [0.0, 1.0]])
    xcentres = (xedges[:-1] + xedges[1:]) / 2
    ycentres = (yedges[:-1] + yedges[1:]) / 2

    # Make a valid grid to match ydata
    xx, yy = np.meshgrid(xcentres, ycentres, sparse=False, indexing='ij')

    x0 = Parameter('x0', value=mean[0], min=0.0, max=1.0)
    sig_x = Parameter('sig_x', value=0.2, min=0.0, max=0.3)
    y0 = Parameter('y0', value=mean[1], min=0.0, max=1.0)
    sig_y = Parameter('sig_y', value=0.1, min=0.0, max=0.3)
    A = Parameter('A', value=np.mean(ydata), min=0.0)
    x = Variable('x')
    y = Variable('y')
    g = Variable('g')

    model = GradientModel(
        {g: A * Gaussian(x, x0, sig_x) * Gaussian(y, y0, sig_y)})
    fit = Fit(model, x=xx, y=yy, g=ydata)
    fit_result = fit.execute()

    assert fit_result.value(x0) == pytest.approx(np.mean(data[:, 0]), 1e-3)
    assert fit_result.value(y0) == pytest.approx(np.mean(data[:, 1]), 1e-3)
    assert np.abs(fit_result.value(sig_x)) == pytest.approx(np.std(data[:, 0]), 1e-2)
    assert np.abs(fit_result.value(sig_y)) == pytest.approx(np.std(data[:, 1]), 1e-2)
    assert fit_result.r_squared >= 0.96


def test_gaussian_2d_fitting_background():
    """
    Tests fitting to a scalar gaussian function with 2 independent
    variables to data with a background. Added after #149.
    """
    mean = (0.6, 0.4)  # x, y mean 0.6, 0.4
    cov = [[0.2**2, 0], [0, 0.1**2]]
    background = 3.0

    # TODO: Since we bin this data later on in a 100 bins, just evaluate 100
    #       points on a Gaussian, and add an appropriate amount of noise. This
    #       burns CPU cycles without good reason.
    data = np.random.multivariate_normal(mean, cov, 500000)
    # print(data.shape)
    # Insert them as y,x here as np fucks up cartesian conventions.
    ydata, xedges, yedges = np.histogram2d(data[:, 0], data[:, 1], bins=100,
                                           range=[[0.0, 1.0], [0.0, 1.0]])
    xcentres = (xedges[:-1] + xedges[1:]) / 2
    ycentres = (yedges[:-1] + yedges[1:]) / 2
    ydata += background  # Background

    # Make a valid grid to match ydata
    xx, yy = np.meshgrid(xcentres, ycentres, sparse=False, indexing='ij')

    x0 = Parameter('x0', value=1.1 * mean[0], min=0.0, max=1.0)
    sig_x = Parameter('sig_x', value=1.1 * 0.2, min=0.0, max=0.3)
    y0 = Parameter('y0', value=1.1 * mean[1], min=0.0, max=1.0)
    sig_y = Parameter('sig_y', value=1.1 * 0.1, min=0.0, max=0.3)
    A = Parameter('A', value=1.1 * np.mean(ydata), min=0.0)
    b = Parameter('b', value=1.2 * background, min=0.0)
    x = Variable('x')
    y = Variable('y')
    g = Variable('g')

    model = GradientModel({g: A * Gaussian(x, x0, sig_x) * Gaussian(y, y0, sig_y) + b})

    # ydata, = model(x=xx, y=yy, x0=mean[0], y0=mean[1], sig_x=np.sqrt(cov[0][0]), sig_y=np.sqrt(cov[1][1]), A=1, b=3.0)
    fit = Fit(model, x=xx, y=yy, g=ydata)
    fit_result = fit.execute()

    assert fit_result.value(x0) / np.mean(data[:, 0]) == pytest.approx(1.0, 1e-2)
    assert fit_result.value(y0) / np.mean(data[:, 1]) == pytest.approx(1.0, 1e-2)
    assert np.abs(fit_result.value(sig_x)) / np.std(data[:, 0]) == pytest.approx(1.0, 1e-2)
    assert np.abs(fit_result.value(sig_y)) / np.std(data[:, 1]) == pytest.approx(1.0, 1e-2)
    assert background / fit_result.value(b) == pytest.approx(1.0, 1e-1)
    assert fit_result.r_squared >= 0.96



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from symfit import Parameter, Variable, Fit, GradientModel
from symfit.distributions import Gaussian

palette = sns.color_palette()


x = Variable('x')
y = Variable('y')
A = Parameter('A')
sig = Parameter(name='sig', value=1.4, min=1.0, max=2.0)
x0 = Parameter(name='x0', value=15.0, min=0.0)

# Gaussian distribution
model = GradientModel({y: A*Gaussian(x, x0, sig)})

# Sample 10000 points from a N(15.0, 1.5) distrubution
np.random.seed(seed=123456789)
sample = np.random.normal(loc=15.0, scale=1.5, size=(10000,))
ydata, bin_edges = np.histogram(sample, 100)
xdata = (bin_edges[1:] + bin_edges[:-1])/2

fit = Fit(model, xdata, ydata)
fit_result = fit.execute()

y, = model(x=xdata, **fit_result.params)
sns.regplot(xdata, ydata, fit_reg=False)
plt.plot(xdata, y, color=palette[2])
plt.ylim(0, 400)
plt.show()



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

from __future__ import division, print_function
import pickle
import pytest
import sympy

from symfit import Variable, Parameter


def test_parameter_add():
    """
    Makes sure the __add__ method of Parameters behaves as expected.
    """
    a = Parameter('a', value=1.0, min=0.5, max=1.5)
    b = Parameter('b', value=1.0, min=0.0)
    new = a + b
    assert isinstance(new, sympy.Add)


def test_argument_unnamed():
    """
    Make sure the generated parameter names follow the pattern
    """
    with pytest.warns(DeprecationWarning):
        a = Parameter()
    b = Parameter('b', 10)
    with pytest.warns(DeprecationWarning):
        c = Parameter(value=10)
    with pytest.warns(DeprecationWarning):
        x = Variable()
    y = Variable('y')

    assert str(a) == '{}_{}'.format(a._argument_name, a._argument_index)
    assert str(a) == 'par_{}'.format(a._argument_index)
    assert str(b) != '{}_{}'.format(b._argument_name, b._argument_index)
    assert str(c) == '{}_{}'.format(c._argument_name, c._argument_index)
    assert c.value == 10
    assert b.value == 10
    assert str(x) == 'var_{}'.format(x._argument_index)
    assert str(y) == 'y'

    with pytest.raises(TypeError):
        d = Parameter(10)


def test_pickle():
    """
    Make sure attributes are preserved when pickling
    """
    A = Parameter('A', min=0., max=1e3, fixed=True)
    new_A = pickle.loads(pickle.dumps(A))
    assert (A.min, A.value, A.max, A.fixed, A.name) == (new_A.min, new_A.value, new_A.max, new_A.fixed, new_A.name)

    with pytest.warns(DeprecationWarning):
        A = Parameter(min=0., max=1e3, fixed=True)
    new_A = pickle.loads(pickle.dumps(A))
    assert (A.min, A.value, A.max, A.fixed, A.name) == (new_A.min, new_A.value, new_A.max, new_A.fixed, new_A.name)


def test_slots():
    """
    Make sure Parameters and Variables don't have a __dict__
    """
    P = Parameter('P')

    # If you only have __slots__ you can't set arbitrary attributes, but
    # you *should* be able to set those that are in your __slots__
    try:
        P.min = 0
    except AttributeError:
        assert False

    with pytest.raises(AttributeError):
        P.foo = None

    V = Variable('V')
    with pytest.raises(AttributeError):
        V.bar = None



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

# -*- coding: utf-8 -*-
from __future__ import division, print_function
import pytest

import numpy as np

from symfit import (
    Fit, Parameter, Variable, Model, GradientModel
)
from symfit.core.minimizers import BFGS, DifferentialEvolution
from symfit.distributions import Gaussian

import inspect


class TestGlobalOptGaussian:
    @classmethod
    def setup_class(cls):
        np.random.seed(0)
        mean = (0.4, 0.4)  # x, y mean 0.6, 0.4
        cov = [[0.01**2, 0], [0, 0.01**2]]

        # TODO: evaluate gaussian at 200x200 points (?!) and add appropriate noise
        data = np.random.multivariate_normal(mean, cov, 2500000)

        # Insert them as y,x here as np fucks up cartesian conventions.
        cls.ydata, xedges, yedges = np.histogram2d(data[:, 1], data[:, 0], bins=200,
                                                   range=[[0.0, 1.0], [0.0, 1.0]])
        xcentres = (xedges[:-1] + xedges[1:]) / 2
        ycentres = (yedges[:-1] + yedges[1:]) / 2

        # Make a valid grid to match ydata
        cls.xx, cls.yy = np.meshgrid(xcentres, ycentres, sparse=False)
        # xdata = np.dstack((xx, yy)).T

    def setup_method(self):
        x = Variable('x')
        y = Variable('y')
        xmin, xmax = -5, 5
        self.x0_1 = Parameter('x01', value=0, min=xmin, max=xmax)
        self.sig_x_1 = Parameter('sigx1', value=0, min=0.0, max=1)
        self.y0_1 = Parameter('y01', value=0, min=xmin, max=xmax)
        self.sig_y_1 = Parameter('sigy1', value=0, min=0.0, max=1)
        self.A_1 = Parameter('A1', min=0, max=1000)
        g_1 = self.A_1 * Gaussian(x, self.x0_1, self.sig_x_1) * Gaussian(y, self.y0_1, self.sig_y_1)

        self.model = GradientModel(g_1)

    def test_diff_evo(self):
        """
        Tests fitting to a scalar gaussian with 2 independent variables with
        wide bounds.
        """

        fit = Fit(self.model, self.xx, self.yy, self.ydata, minimizer=BFGS)
        fit_result = fit.execute()

        assert isinstance(fit.minimizer, BFGS)

        # Make sure a local optimizer doesn't find the answer.
        assert not fit_result.value(self.x0_1) == pytest.approx(0.4, 1e-1)
        assert not fit_result.value(self.y0_1) == pytest.approx(0.4, 1e-1)

        # On to the main event
        fit = Fit(self.model, self.xx, self.yy, self.ydata,
                  minimizer=DifferentialEvolution)
        fit_result = fit.execute(polish=True, seed=0, tol=1e-4, maxiter=100)
        # Global minimizers are really bad at finding local minima though, so
        # roughly equal is good enough.
        assert fit_result.value(self.x0_1) == pytest.approx(0.4, 1e-1)
        assert fit_result.value(self.y0_1) == pytest.approx(0.4, 1e-1)

    def test_chained_min(self):
        """Test fitting with a chained minimizer"""
        curvals = [p.value for p in self.model.params]
        fit = Fit(self.model, self.xx, self.yy, self.ydata,
                  minimizer=[DifferentialEvolution, BFGS])
        fit_result = fit.execute(
            DifferentialEvolution={'seed': 0, 'tol': 1e-4, 'maxiter': 10}
        )
        assert fit_result.value(self.x0_1) == pytest.approx(0.4, 1e-4)
        assert fit_result.value(self.y0_1) == pytest.approx(0.4, 1e-4)
        assert curvals == [p.value for p in self.model.params]

    def test_chained_min_signature(self):
        """
        Test the automatic generation of the signature for ChainedMinimizer
        """
        minimizers = [
            BFGS, DifferentialEvolution, BFGS, DifferentialEvolution, BFGS
        ]

        fit = Fit(self.model, self.xx, self.yy, self.ydata,
                  minimizer=minimizers)

        names = [
            'BFGS', 'DifferentialEvolution', 'BFGS_2',
            'DifferentialEvolution_2', 'BFGS_3'
        ]
        for name, param_name in zip(names, fit.minimizer.__signature__.parameters):
            assert name == param_name
        # Check for equal lengths because zip is slippery that way
        assert len(names) == len(fit.minimizer.__signature__.parameters)

        for param in fit.minimizer.__signature__.parameters.values():
            assert param.kind == inspect.Parameter.KEYWORD_ONLY
        # Make sure keywords end up at the right minimizer.
        with pytest.raises(TypeError):
            # This is not a valid kwarg to DiffEvo, but it is to BFGS. Check if
            # we really go by name of the Minimizer, not by order.
            fit.execute(DifferentialEvolution={'return_all': False})


def test_mexican_hat():
    """
    Test that global minimisation finds the global minima, and doesn't
    affect the value of parameters.
    """
    x = Parameter('x')
    x.min, x.max = -100, 100
    x.value = -2.5
    y = Variable('y')

    model = Model({y: x**4 - 10 * x**2 - x})  # Skewed Mexican hat
    fit = Fit(model, minimizer=[DifferentialEvolution, BFGS])
    fit_result1 = fit.execute(DifferentialEvolution={'seed': 0})

    fit = Fit(model)
    fit_result2 = fit.execute()

    assert fit_result1.value(x) > 0
    assert fit_result2.value(x) < 0



from symfit import variables, parameters, Fit, D, ODEModel
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# Example of the easy of use of the symfit ODE integration syntax.
a, b, c, d, t = variables('a, b, c, d, t')
k, p, l, m = parameters('k, p, l, m')

a0 = 10
b = a0 - d + a # [B] is not independent.

model_dict = {
    D(d, t): l * c * b - m * d,
    D(c, t): k * a * b - p * c - l * c * b + m * d,
    D(a, t): - k * a * b + p * c,
}

model = ODEModel(model_dict, initial={t: 0.0, a: a0, c: 0.0, d: 0.0})

# Generate some data
tdata = np.linspace(0, 3, 1000)
# Eval the normal way.
AA, AAB, BAAB = model(t=tdata, k=0.1, l=0.2, m=.3, p=0.3)

plt.plot(tdata, AA, color='red', label='[AA]')
plt.plot(tdata, AAB, color='blue', label='[AAB]')
plt.plot(tdata, BAAB, color='green', label='[BAAB]')
plt.plot(tdata, b(d=BAAB, a=AA), color='pink', label='[B]')
# plt.plot(tdata, AA + AAB + BAAB, color='black', label='total')
plt.legend()
plt.show()


# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

from __future__ import division, print_function
import pytest
import warnings

import numpy as np
import pickle
import multiprocessing as mp

from symfit import (
    Variable, Parameter, Eq, Ge, parameters, Fit,
    Model, FitResults, variables, CallableNumericalModel
)
from symfit.core.minimizers import *
from symfit.core.objectives import LeastSquares, MinimizeModel, VectorLeastSquares

# Defined at the global level because local functions can't be pickled.


def f(x, a, b):
    return a * x + b


def chi_squared(x, y, a, b, sum=True):
    if sum:
        return np.sum((y - f(x, a, b)) ** 2)
    else:
        return (y - f(x, a, b)) ** 2


def worker(fit_obj):
    return fit_obj.execute()


class SqrtLeastSquares(LeastSquares):
    """
    Minimizes the square root of LeastSquares. This seems to help SLSQP in
    particular, and is considered good practise since the square can grow to
    rapidly, leading to numerical errors.
    """
    # TODO: Make this a standard objective, and perhaps even THE standard
    # objective. This lightweight version is given without proper testing
    # because only the call is relevant, and this makes our multiprocessing test
    # work.

    def __call__(self, *args, **kwargs):
        chi2 = super(SqrtLeastSquares, self).__call__(*args, **kwargs)
        return np.sqrt(chi2)

    def eval_jacobian(self, *args, **kwargs):
        sqrt_chi2 = self(*args, **kwargs)
        chi2_jac = super(SqrtLeastSquares, self).eval_jacobian(*args, **kwargs)
        return 0.5 * (1 / sqrt_chi2) * chi2_jac

    def eval_hessian(self, *args, **kwargs):
        sqrt_chi2 = self(*args, **kwargs)
        sqrt_chi2_jac = self.eval_jacobian(*args, **kwargs)
        chi2 = super(SqrtLeastSquares, self).__call__(*args, **kwargs)
        chi2_jac = super(SqrtLeastSquares, self).eval_jacobian(*args, **kwargs)
        chi2_hess = super(SqrtLeastSquares, self).eval_hessian(*args, **kwargs)
        return - 0.5 * (1 / chi2) * np.outer(sqrt_chi2_jac, chi2_jac) + 0.5 * (1 / sqrt_chi2) * chi2_hess


def subclasses(base, leaves_only=True):
    """
    Recursively create a set of subclasses of ``object``.

    :param object: Class
    :param leaves_only: If ``True``, return only the leaves of the subclass tree
    :return: (All leaves of) the subclass tree.
    """
    base_subs = set(base.__subclasses__())
    if not base_subs or not leaves_only:
        all_subs = {base}
    else:
        all_subs = set()
    for sub in list(base_subs):
        sub_subs = subclasses(sub, leaves_only=leaves_only)
        all_subs.update(sub_subs)
    return all_subs


def setup_function():
    np.random.seed(0)


def test_custom_objective(recwarn):
    """
    Compare the result of a custom objective with the symbolic result.
    :return:
    """
    # Create test data
    xdata = np.linspace(0, 100, 25)  # From 0 to 100 in 100 steps
    a_vec = np.random.normal(15.0, scale=2.0, size=xdata.shape)
    b_vec = np.random.normal(100, scale=2.0, size=xdata.shape)
    ydata = a_vec * xdata + b_vec  # Point scattered around the line 5 * x + 105

    # Normal symbolic fit
    a = Parameter('a', value=0, min=0.0, max=1000)
    b = Parameter('b', value=0, min=0.0, max=1000)
    x = Variable('x')
    y = Variable('y')
    model = {y: a * x + b}

    fit = Fit(model, xdata, ydata, minimizer=BFGS)
    fit_result = fit.execute()

    def f(x, a, b):
        return a * x + b

    def chi_squared(a, b):
        return np.sum((ydata - f(xdata, a, b))**2)

    # Should no longer raise warnings, because internally we practice
    # what we preach.
    fit_custom = BFGS(chi_squared, [a, b])
    assert len(recwarn) == 0

    fit_custom_result = fit_custom.execute()

    assert isinstance(fit_custom_result, FitResults)
    assert fit_custom_result.value(a) == pytest.approx(fit_result.value(a), 1e-5)
    assert fit_custom_result.value(b) == pytest.approx(fit_result.value(b), 1e-4)

    # New preferred usage, multi component friendly.
    with pytest.raises(TypeError):
        callable_model = CallableNumericalModel(
            chi_squared,
            connectivity_mapping={y: {a, b}}
        )
    callable_model = CallableNumericalModel(
        {y: chi_squared},
        connectivity_mapping={y: {a, b}}
    )
    assert callable_model.params == [a, b]
    assert callable_model.independent_vars == []
    assert callable_model.dependent_vars == [y]
    assert callable_model.interdependent_vars == []
    assert callable_model.connectivity_mapping == {y: {a, b}}
    fit_custom = BFGS(callable_model, [a, b])
    fit_custom_result = fit_custom.execute()

    assert isinstance(fit_custom_result, FitResults)
    assert fit_custom_result.value(a) == pytest.approx(fit_result.value(a), 1e-5)
    assert fit_custom_result.value(b) == pytest.approx(fit_result.value(b), 1e-4)


def test_custom_parameter_names():
    """
    For cusom objective functions you still have to provide a list of Parameter
    objects to use with the same name as the keyword arguments to your function.
    """
    a = Parameter('a')
    c = Parameter('c')

    def chi_squared(a, b):
        """
        Dummy function with different keyword argument names
        """
        pass

    fit_custom = BFGS(chi_squared, [a, c])
    with pytest.raises(TypeError):
        fit_custom.execute()


def test_powell():
    """
    Powell with a single parameter gave an error because a 0-d array was
    returned by scipy. So no error here is winning.
    """
    x, y = variables('x, y')
    a, b = parameters('a, b')
    b.fixed = True

    model = Model({y: a * x + b})
    xdata = np.linspace(0, 10)
    ydata = model(x=xdata, a=5.5, b=15.0).y + np.random.normal(0, 1)
    fit = Fit({y: a * x + b}, x=xdata, y=ydata, minimizer=Powell)
    fit_result = fit.execute()
    assert fit_result.value(b) == pytest.approx(1.0)


def test_jac_hess():
    """
    Make sure both the Jacobian and Hessian are passed to the minimizer.
    """
    x, y = variables('x, y')
    a, b = parameters('a, b')
    b.fixed = True

    model = Model({y: a * x + b})
    xdata = np.linspace(0, 10)
    ydata = model(x=xdata, a=5.5, b=15.0).y + np.random.normal(0, 1)
    fit = Fit({y: a * x + b}, x=xdata, y=ydata, minimizer=TrustConstr)
    assert isinstance(fit.minimizer.objective, LeastSquares)
    assert isinstance(fit.minimizer.jacobian.__self__, LeastSquares)
    assert isinstance(fit.minimizer.hessian.__self__, LeastSquares)

    fit_result = fit.execute()
    assert fit_result.value(b) == pytest.approx(1.0)


def test_pickle():
    """
    Test the picklability of the different minimizers.
    """
    # Create test data
    xdata = np.linspace(0, 100, 100)  # From 0 to 100 in 100 steps
    a_vec = np.random.normal(15.0, scale=2.0, size=xdata.shape)
    b_vec = np.random.normal(100, scale=2.0, size=xdata.shape)
    ydata = a_vec * xdata + b_vec  # Point scattered around the line 5 * x + 105

    # Normal symbolic fit
    a = Parameter('a', value=0, min=0.0, max=1000)
    b = Parameter('b', value=0, min=0.0, max=1000)
    x, y = variables('x, y')

    # Make a set of all ScipyMinimizers, and add a chained minimizer.
    scipy_minimizers = list(subclasses(ScipyMinimize))
    chained_minimizer = (DifferentialEvolution, BFGS)
    scipy_minimizers.append(chained_minimizer)
    constrained_minimizers = subclasses(ScipyConstrainedMinimize)
    # Test for all of them if they can be pickled.
    for minimizer in scipy_minimizers:
        if minimizer in constrained_minimizers:
            constraints = [Ge(b, a)]
        else:
            constraints = []
        model = CallableNumericalModel(
            {y: f},
            connectivity_mapping={y: {x, a, b}},
        )
        fit = Fit(model, x=xdata, y=ydata, minimizer=minimizer,
                  constraints=constraints)
        if minimizer is not MINPACK:
            assert isinstance(fit.objective, LeastSquares)
            assert isinstance(fit.minimizer.objective, LeastSquares)
        else:
            assert isinstance(fit.objective, VectorLeastSquares)
            assert isinstance(fit.minimizer.objective, VectorLeastSquares)

        fit = fit.minimizer  # Just check if the minimizer pickles
        dump = pickle.dumps(fit)
        pickled_fit = pickle.loads(dump)
        problematic_attr = [
            'objective', '_pickle_kwargs', 'wrapped_objective',
            'constraints', 'wrapped_constraints',
            'local_minimizer', 'minimizers'
        ]

        for key, value in fit.__dict__.items():
            new_value = pickled_fit.__dict__[key]
            try:
                assert value == new_value
            except AssertionError as err:
                if key not in problematic_attr:
                    raise err
                # These attr are new instances, and therefore do not
                # pass an equality test. All we can do is see if they
                # are at least the same type.
                if isinstance(value, (list, tuple)):
                    for val1, val2 in zip(value, new_value):
                        assert isinstance(val1, val2.__class__)
                        if key == 'constraints':
                            assert val1.model.constraint_type == val2.model.constraint_type
                            assert list(val1.model.model_dict.values())[0] == list(val2.model.model_dict.values())[0]
                            assert val1.model.independent_vars == val2.model.independent_vars
                            assert val1.model.params == val2.model.params
                            assert val1.model.__signature__ == val2.model.__signature__
                        elif key == 'wrapped_constraints':
                            if isinstance(val1, dict):
                                assert val1['type'] == val2['type']
                                assert set(val1.keys()) == set(val2.keys())
                            elif isinstance(val1, NonlinearConstraint):
                                # For trust-ncg we manually check if
                                # their dicts are equal, because no
                                # __eq__ is implemented on
                                # NonLinearConstraint
                                assert len(val1.__dict__) == len(val2.__dict__)
                                for key in val1.__dict__:
                                    try:
                                        assert val1.__dict__[key] == val2.__dict__[key]
                                    except AssertionError:
                                        assert isinstance(val1.__dict__[key], val2.__dict__[key].__class__)
                            else:
                                raise NotImplementedError('No such constraint type is known.')
                elif key == '_pickle_kwargs':
                    FitResults._array_safe_dict_eq(value, new_value)
                else:
                    assert isinstance(new_value, value.__class__)
        assert set(fit.__dict__.keys()) == set(pickled_fit.__dict__.keys())

        # Test if we converge to the same result.
        np.random.seed(2)
        res_before = fit.execute()
        np.random.seed(2)
        res_after = pickled_fit.execute()
        assert FitResults._array_safe_dict_eq(res_before.__dict__, res_after.__dict__)


def test_multiprocessing():
    """
    To make sure pickling truly works, try multiprocessing. No news is good
    news.
    """
    np.random.seed(2)
    x = np.arange(100, dtype=float)
    a_values = np.array([1, 2, 3])
    np.random.shuffle(a_values)

    def gen_fit_objs(x, a, minimizer):
        """Generates linear fits with different a parameter values."""
        for a_i in a:
            a_par = Parameter('a', 4.0, min=0.0, max=20)
            b_par = Parameter('b', 1.2, min=0.0, max=2)
            x_var = Variable('x')
            y_var = Variable('y')

            con_map = {y_var: {x_var, a_par, b_par}}
            model = CallableNumericalModel({y_var: f}, connectivity_mapping=con_map)

            fit = Fit(
                model, x, a_i * x + 1, minimizer=minimizer,
                objective=SqrtLeastSquares if minimizer is not MINPACK else VectorLeastSquares
            )
            yield fit

    minimizers = subclasses(ScipyMinimize)
    chained_minimizer = (DifferentialEvolution, BFGS)
    minimizers.add(chained_minimizer)

    pool = mp.Pool()
    for minimizer in minimizers:
        results = pool.map(worker, gen_fit_objs(x, a_values, minimizer))
        a_results = [res.params['a'] for res in results]
        # Check the results
        assert a_values == pytest.approx(a_results, 1e-2)
        for result in results:
            # Check that we are actually using the right minimizer
            if isinstance(result.minimizer, ChainedMinimizer):
                for used, target in zip(result.minimizer.minimizers, minimizer):
                    assert isinstance(used, target)
            else:
                assert isinstance(result.minimizer, minimizer)
            assert isinstance(result.iterations, int) or result.iterations is None


def test_minimizer_constraint_compatibility():
    """
    Test if #156 has been solved, and test all the other constraint styles.
    """
    x, y, z = variables('x, y, z')
    a, b, c = parameters('a, b, c')
    b.fixed = True

    model = Model({z: a * x**2 - b * y**2 + c})
    # Generate data, z has to be scalar for MinimizeModel to be happy
    xdata = 3  # np.linspace(0, 10)
    ydata = 5  # np.linspace(0, 10)
    zdata = model(a=2, b=3, c=5, x=xdata, y=ydata).z
    data_dict = {x: xdata, y: ydata, z: zdata}

    # Equivalent ways of defining the same constraint
    constraint_model = Model.as_constraint(a - c, model, constraint_type=Eq)
    constraint_model.params = model.params
    constraints = [
        Eq(a, c),
        MinimizeModel(constraint_model, data=data_dict),
        constraint_model
    ]

    objective = MinimizeModel(model, data=data_dict)
    for constraint in constraints:
        fit = SLSQP(objective, parameters=[a, b, c],
                    constraints=[constraint])
        wrapped_constr = fit.wrapped_constraints[0]['fun'].model
        assert isinstance(wrapped_constr, Model)
        assert wrapped_constr.params == model.params
        assert wrapped_constr.jacobian_model.params == model.params
        assert wrapped_constr.hessian_model.params == model.params
        # Set the data for the dependent var of the constraint to None
        # Normally this is handled by Fit because here we interact with the
        # Minimizer directly, it is up to us.
        constraint_var = fit.wrapped_constraints[0]['fun'].model.dependent_vars[0]
        objective.data[constraint_var] = None
        fit.execute()

    # No scipy style dicts allowed.
    with pytest.raises(TypeError):
        fit = SLSQP(MinimizeModel(model, data=data_dict),
                    parameters=[a, b, c],
                    constraints=[{'type': 'eq', 'fun': lambda a, b, c: a - c}])



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

# -*- coding: utf-8 -*-
from symfit import Variable, Parameter, Fit, Model
from symfit.contrib.interactive_guess import InteractiveGuess
import numpy as np


x = Variable('x')
y1 = Variable('y1')
y2 = Variable('y2')
k = Parameter('k', 900)
x0 = Parameter('x0', 1.5)

model = {
    y1: k * (x-x0)**2,
    y2: x - x0
}
model = Model(model)

# Generate example data
x_data = np.linspace(0, 2.5, 50)
data = model(x=x_data, k=1000, x0=1)
y1_data = data.y1
y2_data = data.y2

guess = InteractiveGuess(model, x=x_data, y1=y1_data, y2=y2_data, n_points=250)
guess.execute()
print(guess)

fit = Fit(model, x=x_data, y1=y1_data, y2=y2_data)
fit_result = fit.execute()
print(fit_result)



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

"""
``symfit`` occasionally updates the printing of ``sympy`` objects, such that they
print into their ``numpy``/``scipy`` equivalent. This is done because sometimes
such printing has not been implemented in ``sympy`` yet, or because we want
slightly different behavior from the standard one.

Users using both ``symfit`` and ``sympy`` should be aware of this.
"""
import pkg_resources
from sympy import HadamardProduct, MatPow, Idx, Inverse
from sympy.printing.codeprinter import CodePrinter

##########################################################
# Monkeypatch the printers until this is merged upstream #
##########################################################

class DontDeleteMe(object):
    def __init__(self, default_value):
        self.dont_delete = default_value
        self.default_value = default_value

    def __get__(self, instance, owner):
        return self.dont_delete

    def __set__(self, instance, value):
        self.dont_delete = value

    def __delete__(self, instance):
        self.dont_delete = self.default_value

CodePrinter._not_supported = DontDeleteMe(set())
CodePrinter._number_symbols = DontDeleteMe(set())

def _print_Inverse(self, printer):
    return "%s(%s)" % (printer._module_format('numpy.linalg.inv'),
                       printer.doprint(self.args[0]))
Inverse._numpycode = _print_Inverse

def _print_HadamardProduct(self, printer):
    return "%s*%s" % (printer.doprint(self.args[0]),
                      printer.doprint(self.args[1]))
HadamardProduct._numpycode = _print_HadamardProduct

if pkg_resources.parse_version(pkg_resources.get_distribution('sympy').version) \
        > pkg_resources.parse_version('1.4'):
    from sympy import HadamardPower

    def _print_HadamardPower(self, printer):
        return "%s**%s" % (printer.doprint(self.args[0]),
                          printer.doprint(self.args[1]))
    HadamardPower._numpycode = _print_HadamardPower

def _print_Idx(self, printer):
    """
    Print ``Idx`` objects.
    """
    return "{0}".format(printer.doprint(self.args[0]))
Idx._numpycode = _print_Idx

def _print_MatPow(self, printer):
    if self.shape == (1, 1):
        # Scalar, so we can take a normal power.
        return printer._print_Pow(self)
    else:
        return printer._print_MatPow(self)
MatPow._numpycode = _print_MatPow

#########################################################
# End of monkeypatch                                    #
#########################################################


# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

from __future__ import division, print_function

import numpy as np
import scipy.stats
from scipy.optimize import curve_fit
import pytest

from symfit import (
    Variable, Parameter, Fit, FitResults, log, variables,
    parameters, Model, exp, integrate, oo, GradientModel
)
from symfit.core.minimizers import (
    MINPACK, LBFGSB, BoundedMinimizer, DifferentialEvolution, BaseMinimizer,
    ChainedMinimizer
)
from symfit.core.objectives import LogLikelihood, MinimizeModel, LeastSquares
from symfit.distributions import Gaussian, Exp, BivariateGaussian
from tests.test_minimizers import subclasses

import inspect


def setup_module():
    np.random.seed(0)


def test_callable():
    """
    Make sure that symfit expressions are callable (with scalars and
    arrays), and produce the expected results.
    """
    a, b = parameters('a, b')
    x, y = variables('x, y')
    func = a*x**2 + b*y**2
    result = func(x=2, y=3, a=3, b=9)
    assert result == 3*2**2 + 9*3**2
    result = func(2, 3, a=3, b=9)
    assert result == 3*2**2 + 9*3**2

    xdata = np.arange(1, 10)
    ydata = np.arange(1, 10)
    result = func(x=ydata, y=ydata, a=3, b=9)
    assert np.array_equal(result, 3*xdata**2 + 9*ydata**2)


def test_named_fitting():
    """
    Make sure that fitting with NumericalLeastSquares works using a dict
    as model and that the resulting fit_result is of the right type.
    """
    xdata = np.linspace(1, 10, 10)
    ydata = 3*xdata**2

    a = Parameter('a', value=1.0)
    b = Parameter('b', value=2.5)
    x, y = variables('x, y')

    model = {y: a*x**b}

    fit = Fit(model, x=xdata, y=ydata, minimizer=MINPACK)
    fit_result = fit.execute()
    assert isinstance(fit_result, FitResults)
    assert fit_result.value(a) == pytest.approx(3.0)
    assert fit_result.value(b) == pytest.approx(2.0)


def test_backwards_compatible_fitting():
    """
    In 0.4.2 we replaced the usage of inspect by automatically generated
    names. This can cause problems for users using named variables to call
    fit.
    """
    xdata = np.linspace(1, 10, 10)
    ydata = 3*xdata**2

    a = Parameter('a', value=1.0)
    b = Parameter('b', value=2.5)

    y = Variable('y')

    with pytest.warns(DeprecationWarning):
        x = Variable()

    model = {y: a*x**b}

    with pytest.raises(TypeError):
        fit = Fit(model, x=xdata, y=ydata)


def test_vector_fitting():
    """
    Tests fitting to a 3 component vector valued function, without bounds
    or guesses.
    """
    a, b, c = parameters('a, b, c')
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = {a_i: a, b_i: b, c_i: c}

    xdata = np.array([
        [10.1, 9., 10.5, 11.2, 9.5, 9.6, 10.],
        [102.1, 101., 100.4, 100.8, 99.2, 100., 100.8],
        [71.6, 73.2, 69.5, 70.2, 70.8, 70.6, 70.1],
    ])

    fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
        minimizer=MINPACK
    )
    fit_result = fit.execute()

    assert fit_result.value(a) == pytest.approx(np.mean(xdata[0]), 1e-5)
    assert fit_result.value(b) == pytest.approx(np.mean(xdata[1]), 1e-4)
    assert fit_result.value(c) == pytest.approx(np.mean(xdata[2]), 1e-5)


def test_vector_none_fitting():
    """
    Fit to a 3 component vector valued function with one variables data set
    to None, without bounds or guesses.
    """
    a, b, c = parameters('a, b, c')
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = {a_i: a, b_i: b, c_i: c}

    xdata = np.array([
        [10.1, 9., 10.5, 11.2, 9.5, 9.6, 10.],
        [102.1, 101., 100.4, 100.8, 99.2, 100., 100.8],
        [71.6, 73.2, 69.5, 70.2, 70.8, 70.6, 70.1],
    ])

    fit_none = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=None,
        minimizer=MINPACK
    )
    fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
        minimizer=MINPACK
    )
    fit_none_result = fit_none.execute()
    fit_result = fit.execute()

    assert fit_none_result.value(b) == pytest.approx(fit_result.value(b), 1e-4)
    assert fit_none_result.value(a) == pytest.approx(fit_result.value(a), 1e-4)
    # the parameter without data should be unchanged.
    assert fit_none_result.value(c) == pytest.approx(1.0)


def test_vector_fitting_guess():
    """
    Tests fitting to a 3 component vector valued function, with guesses.
    """
    a, b, c = parameters('a, b, c')
    a.value = 10
    b.value = 100
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = {a_i: a, b_i: b, c_i: c}

    xdata = np.array([
        [10.1, 9., 10.5, 11.2, 9.5, 9.6, 10.],
        [102.1, 101., 100.4, 100.8, 99.2, 100., 100.8],
        [71.6, 73.2, 69.5, 70.2, 70.8, 70.6, 70.1],
    ])

    fit = Fit(
        model=model,
        a_i=xdata[0],
        b_i=xdata[1],
        c_i=xdata[2],
        minimizer=MINPACK
    )
    fit_result = fit.execute()

    assert fit_result.value(a) == pytest.approx(np.mean(xdata[0]), 1e-4)
    assert fit_result.value(b) == pytest.approx(np.mean(xdata[1]), 1e-4)
    assert fit_result.value(c) == pytest.approx(np.mean(xdata[2]), 1e-4)


def test_fitting():
    """
    Tests fitting with NumericalLeastSquares. Makes sure that the resulting
    objects and values are of the right type, and that the fit_result does
    not have unexpected members.
    """
    xdata = np.linspace(1, 10, 10)
    ydata = 3*xdata**2

    a = Parameter('a')  # 3.1, min=2.5, max=3.5
    b = Parameter('b')
    x = Variable('x')
    new = a*x**b

    fit = Fit(new, xdata, ydata, minimizer=MINPACK)

    fit_result = fit.execute()
    assert isinstance(fit_result, FitResults)
    assert fit_result.value(a) == pytest.approx(3.0)
    assert fit_result.value(b) == pytest.approx(2.0)

    assert isinstance(fit_result.stdev(a), float)
    assert isinstance(fit_result.stdev(b), float)

    assert isinstance(fit_result.r_squared, float)
    assert fit_result.r_squared == 1.0  # by definition since there's no fuzzyness


def test_grid_fitting():
    """
    Tests fitting a scalar function with 2 independent variables.
    """
    xdata = np.arange(-5, 5, 1)
    ydata = np.arange(5, 15, 1)
    xx, yy = np.meshgrid(xdata, ydata, sparse=False)

    zdata = (2.5*xx**2 + 3.0*yy**2)

    a = Parameter('a', value=2.5, max=2.75)
    b = Parameter('b', value=3.0, min=2.75)
    x = Variable('x')
    y = Variable('y')
    z = Variable('z')
    new = {z: a*x**2 + b*y**2}

    fit = Fit(new, x=xx, y=yy, z=zdata)
    results = fit.execute()

    assert isinstance(fit.minimizer, LBFGSB)

    assert results.value(a) == pytest.approx(2.5)
    assert results.value(b) == pytest.approx(3.)


def test_model_callable():
    """
    Tests if Model objects are callable in the way expected. Calling a
    model should evaluate it's expression(s) with the given values. The
    return value is a namedtuple.

    The signature should also work so inspection is saved.
    """
    a, b = parameters('a, b')
    x, y = variables('x, y')
    new = a*x**2 + b*y**2
    model = Model(new)
    ans = model(3, 3, 2, 2)
    assert isinstance(ans, tuple)
    z, = ans

    assert z == 36
    for arg_name, name in zip(('x', 'y', 'a', 'b'), inspect.signature(model).parameters):
        assert arg_name == name


@pytest.mark.skip(reason=("Test fails a proportion of the time because `z_1`, "
                          "`z_2` and `z_3` are not necessarily ordered as "
                          "expected so the assert statements fail. `z_1` is "
                          "frequenty equal to either 36 or 72."))
def test_model_callable_from_model_init():
    """
    Tests if Model objects are callable in the way expected. Calling a
    model should evaluate it's expression(s) with the given values. The
    return value is a namedtuple.

    The signature should also work so inspection is saved.

    Tests from Model __init__ directly
    """
    a, b = parameters('a, b')
    x, y = variables('x, y')
    model = Model([
        a*x**2,
        4*b*y**2,
        a*x**2 + b*y**2
    ])
    z_1, z_2, z_3 = model(3, 3, 2, 2)

    assert z_1 == 18
    assert z_2 == 72
    assert z_3 == 36
    for arg_name, name in zip(('x', 'y', 'a', 'b'), inspect.signature(model).parameters):
        assert arg_name == name


def test_model_callable_from_dict():
    """
    Tests if Model objects are callable in the way expected. Calling a
    model should evaluate it's expression(s) with the given values. The
    return value is a namedtuple.

    The signature should also work so inspection is saved.

    Tests from dict.
    """
    a, b = parameters('a, b')
    x, y = variables('x, y')
    z_1, z_2, z_3 = variables('z_1, z_2, z_3')
    model = Model({
        z_1: a*x**2,
        z_2: 4*b*y**2,
        z_3: a*x**2 + b*y**2
    })
    z_1, z_2, z_3 = model(3, 3, 2, 2)

    assert z_1 == 18
    assert z_2 == 72
    assert z_3 == 36
    for arg_name, name in zip(('x', 'y', 'a', 'b'), inspect.signature(model).parameters):
        assert arg_name == name


def test_2D_fitting():
    """
    Makes sure that a scalar model with 2 independent variables has the
    proper signature, and that the fit result is of the correct type.
    """
    xdata = np.random.randint(-10, 11, size=(2, 400))
    zdata = 2.5*xdata[0]**2 + 7.0*xdata[1]**2

    a = Parameter('a')
    b = Parameter('b')
    x = Variable('x')
    y = Variable('y')
    new = a*x**2 + b*y**2

    fit = Fit(new, xdata[0], xdata[1], zdata)

    result = fit.model(xdata[0], xdata[1], 2, 3)
    assert isinstance(result, tuple)

    for arg_name, name in zip(('x', 'y', 'a', 'b'), inspect.signature(fit.model).parameters):
        assert arg_name == name

    fit_result = fit.execute()
    assert isinstance(fit_result, FitResults)


def test_gaussian_fitting():
    """
    Tests fitting to a gaussian function and fit_result.params unpacking.
    """
    xdata = 2*np.random.rand(10000) - 1  # random betwen [-1, 1]
    ydata = 5.0 * scipy.stats.norm.pdf(xdata, loc=0.0, scale=1.0)

    x0 = Parameter('x0')
    sig = Parameter('sig')
    A = Parameter('A')
    x = Variable('x')
    g = GradientModel(A * Gaussian(x, x0, sig))

    fit = Fit(g, xdata, ydata)
    assert isinstance(fit.objective, LeastSquares)
    fit_result = fit.execute()

    assert fit_result.value(A) == pytest.approx(5.0)
    assert np.abs(fit_result.value(sig)) == pytest.approx(1.0)
    assert fit_result.value(x0) == pytest.approx(0.0)
    # raise Exception([i for i in fit_result.params])
    sexy = g(x=2.0, **fit_result.params)
    ugly = g(
        x=2.0,
        x0=fit_result.value(x0),
        A=fit_result.value(A),
        sig=fit_result.value(sig),
    )
    assert sexy == ugly


def test_2_gaussian_2d_fitting():
    """
    Tests fitting to a scalar gaussian with 2 independent variables with
    tight bounds.
    """
    mean = (0.3, 0.4)  # x, y mean 0.6, 0.4
    cov = [[0.01**2, 0], [0, 0.01**2]]
    # TODO: evaluate gaussian at 100x100 points and add appropriate noise
    data = np.random.multivariate_normal(mean, cov, 3000000)
    mean = (0.7, 0.8)  # x, y mean 0.6, 0.4
    cov = [[0.01**2, 0], [0, 0.01**2]]
    data_2 = np.random.multivariate_normal(mean, cov, 3000000)
    data = np.vstack((data, data_2))

    # Insert them as y,x here as np fucks up cartesian conventions.
    ydata, xedges, yedges = np.histogram2d(data[:, 1], data[:, 0], bins=100,
                                           range=[[0.0, 1.0], [0.0, 1.0]])
    xcentres = (xedges[:-1] + xedges[1:]) / 2
    ycentres = (yedges[:-1] + yedges[1:]) / 2

    # Make a valid grid to match ydata
    xx, yy = np.meshgrid(xcentres, ycentres, sparse=False)
    # xdata = np.dstack((xx, yy)).T

    x = Variable('x')
    y = Variable('y')

    x0_1 = Parameter('x0_1', value=0.7, min=0.6, max=0.9)
    sig_x_1 = Parameter('sig_x_1', value=0.1, min=0.0, max=0.2)
    y0_1 = Parameter('y0_1', value=0.8, min=0.6, max=0.9)
    sig_y_1 = Parameter('sig_y_1', value=0.1, min=0.0, max=0.2)
    A_1 = Parameter('A_1')
    g_1 = A_1 * Gaussian(x, x0_1, sig_x_1) * Gaussian(y, y0_1, sig_y_1)

    x0_2 = Parameter('x0_2', value=0.3, min=0.2, max=0.5)
    sig_x_2 = Parameter('sig_x_2', value=0.1, min=0.0, max=0.2)
    y0_2 = Parameter('y0_2', value=0.4, min=0.2, max=0.5)
    sig_y_2 = Parameter('sig_y_2', value=0.1, min=0.0, max=0.2)
    A_2 = Parameter('A_2')
    g_2 = A_2 * Gaussian(x, x0_2, sig_x_2) * Gaussian(y, y0_2, sig_y_2)

    model = GradientModel(g_1 + g_2)
    fit = Fit(model, xx, yy, ydata)
    fit_result = fit.execute()

    assert isinstance(fit.minimizer, LBFGSB)

    img = model(x=xx, y=yy, **fit_result.params)[0]
    img_g_1 = g_1(x=xx, y=yy, **fit_result.params)
    img_g_2 = g_2(x=xx, y=yy, **fit_result.params)
    assert img == pytest.approx(img_g_1 + img_g_2)

    # Equal up to some precision. Not much obviously.
    assert fit_result.value(x0_1) == pytest.approx(0.7, 1e-3)
    assert fit_result.value(y0_1) == pytest.approx(0.8, 1e-3)
    assert fit_result.value(x0_2) == pytest.approx(0.3, 1e-3)
    assert fit_result.value(y0_2) == pytest.approx(0.4, 1e-3)


def test_gaussian_2d_fitting():
    """
    Tests fitting to a scalar gaussian function with 2 independent
    variables.
    """
    mean = (0.6, 0.4)  # x, y mean 0.6, 0.4
    cov = [[0.2**2, 0], [0, 0.1**2]]

    # TODO: evaluate gaussian at 100x100 points and add appropriate noise
    data = np.random.multivariate_normal(mean, cov, 1000000)

    # Insert them as y,x here as np fucks up cartesian conventions.
    ydata, xedges, yedges = np.histogram2d(data[:, 0], data[:, 1], bins=100,
                                           range=[[0.0, 1.0], [0.0, 1.0]])
    xcentres = (xedges[:-1] + xedges[1:]) / 2
    ycentres = (yedges[:-1] + yedges[1:]) / 2

    # Make a valid grid to match ydata
    xx, yy = np.meshgrid(xcentres, ycentres, sparse=False, indexing='ij')

    x0 = Parameter('x0', value=mean[0])
    sig_x = Parameter('sig_x', min=0.0)
    x = Variable('x')
    y0 = Parameter('y0', value=mean[1])
    sig_y = Parameter('sig_y', min=0.0)
    A = Parameter('A', min=1, value=100)
    y = Variable('y')
    g = Variable('g')
    # g = A * Gaussian(x, x0, sig_x) * Gaussian(y, y0, sig_y)
    model = Model({g: A * Gaussian(x, x0, sig_x) * Gaussian(y, y0, sig_y)})
    fit = Fit(model, x=xx, y=yy, g=ydata, minimizer=MINPACK)
    fit_result = fit.execute()

    assert fit_result.value(x0) == pytest.approx(np.mean(data[:, 0]), 1e-1)
    assert fit_result.value(y0) == pytest.approx(np.mean(data[:, 1]), 1e-1)
    assert np.abs(fit_result.value(sig_x)) == pytest.approx(np.std(data[:, 0]), 1e-1)
    assert np.abs(fit_result.value(sig_y)) == pytest.approx(np.std(data[:, 1]), 1e-1)
    assert fit_result.r_squared >= 0.99


def test_jacobian_matrix():
    """
    The jacobian matrix of a model should be a 2D list (matrix) containing
    all the partial derivatives.
    """
    a, b, c = parameters('a, b, c')
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = Model({a_i: 2 * a + 3 * b, b_i: 5 * b, c_i: 7 * c})
    assert [[2, 3, 0], [0, 5, 0], [0, 0, 7]] == model.jacobian


def test_hessian_matrix():
    """
    The Hessian matrix of a model should be a 3D list (matrix) containing
    all the 2nd partial derivatives.
    """
    a, b, c = parameters('a, b, c')
    a_i, b_i, c_i = variables('a_i, b_i, c_i')

    model = Model({a_i: 2 * a**2 + 3 * b, b_i: 5 * b**2, c_i: 7 * c*b})
    assert [[[4, 0, 0], [0, 0, 0], [0, 0, 0]],
            [[0, 0, 0], [0, 10, 0], [0, 0, 0]],
            [[0, 0, 0], [0, 0, 7], [0, 7, 0]]] == model.hessian


def test_likelihood_fitting_exponential():
    """
    Fit using the likelihood method.
    """
    b = Parameter('b', value=4, min=3.0)
    x, y = variables('x, y')
    pdf = {y: Exp(x, 1/b)}

    # Draw points from an Exp(5) exponential distribution.
    np.random.seed(100)
    # TODO: Do we *really* need 1m points?
    xdata = np.random.exponential(5, 1000000)

    # Expected parameter values
    mean = np.mean(xdata)
    stdev = np.std(xdata)
    mean_stdev = stdev / np.sqrt(len(xdata))

    with pytest.raises(TypeError):
        fit = Fit(pdf, x=xdata, sigma_y=2.0, objective=LogLikelihood)

    fit = Fit(pdf, xdata, objective=LogLikelihood)
    fit_result = fit.execute()
    pdf_i = fit.model(x=xdata, **fit_result.params).y  # probabilities
    likelihood = np.product(pdf_i)
    loglikelihood = np.sum(np.log(pdf_i))

    assert fit_result.value(b) == pytest.approx(mean, 1e-3)
    assert fit_result.value(b) == pytest.approx(mean, 1e-3)
    assert fit_result.value(b) == pytest.approx(mean, 1e-3)
    assert fit_result.value(b) == pytest.approx(mean, 1e-3)
    assert fit_result.value(b) == pytest.approx(mean, 1e-3)
    assert fit_result.value(b) == pytest.approx(mean, 1e-3)
    assert fit_result.value(b) == pytest.approx(stdev, 1e-3)
    assert fit_result.stdev(b) == pytest.approx(mean_stdev, 1e-3)

    assert likelihood == pytest.approx(fit_result.likelihood)
    assert loglikelihood == pytest.approx(fit_result.log_likelihood)


def test_likelihood_fitting_gaussian():
    """
    Fit using the likelihood method.
    """
    mu, sig = parameters('mu, sig')
    sig.min = 0.01
    sig.value = 3.0
    mu.value = 50.
    x = Variable('x')
    pdf = GradientModel(Gaussian(x, mu, sig))

    np.random.seed(10)
    # TODO: Do we really need 1k points?
    xdata = np.random.normal(51., 3.5, 10000)

    # Expected parameter values
    mean = np.mean(xdata)
    stdev = np.std(xdata)
    mean_stdev = stdev/np.sqrt(len(xdata))

    fit = Fit(pdf, xdata, objective=LogLikelihood)
    fit_result = fit.execute()

    assert fit_result.value(mu) == pytest.approx(mean, 1e-6)
    assert fit_result.stdev(mu) == pytest.approx(mean_stdev, 1e-3)
    assert fit_result.value(sig) == pytest.approx(np.std(xdata), 1e-6)


def test_likelihood_fitting_bivariate_gaussian():
    """
    Fit using the likelihood method.
    """
    # Make variables and parameters
    x = Variable('x')
    y = Variable('y')
    x0 = Parameter('x0', value=0.6, min=0.5, max=0.7)
    sig_x = Parameter('sig_x', value=0.1, max=1.0)
    y0 = Parameter('y0', value=0.7, min=0.6, max=0.9)
    sig_y = Parameter('sig_y', value=0.05, max=1.0)
    rho = Parameter('rho', value=0.001, min=-1, max=1)

    pdf = BivariateGaussian(x=x, mu_x=x0, sig_x=sig_x, y=y, mu_y=y0,
                            sig_y=sig_y, rho=rho)

    # Draw 100000 samples from a bivariate distribution
    mean = [0.59, 0.8]
    r = 0.6
    cov = np.array([[0.11 ** 2, 0.11 * 0.23 * r],
                    [0.11 * 0.23 * r, 0.23 ** 2]])
    np.random.seed(42)
    # TODO: Do we really need 100k points?
    xdata, ydata = np.random.multivariate_normal(mean, cov, 100000).T

    fit = Fit(pdf, x=xdata, y=ydata, objective=LogLikelihood)
    fit_result = fit.execute()

    assert fit_result.value(x0) == pytest.approx(mean[0], 1e-2)
    assert fit_result.value(y0) == pytest.approx(mean[1], 1e-2)
    assert fit_result.value(sig_x) == pytest.approx(np.sqrt(cov[0, 0]), 1e-2)
    assert fit_result.value(sig_y) == pytest.approx(np.sqrt(cov[1, 1]), 1e-2)
    assert fit_result.value(rho) == pytest.approx(r, 1e-2)

    marginal = integrate(pdf, (y, -oo, oo), conds='none')
    fit = Fit(marginal, x=xdata, objective=LogLikelihood)

    with pytest.raises(NameError):    
        # Should raise a NameError, not a TypeError, see #219
        fit.execute()


def test_evaluate_model():
    """
    Makes sure that models are callable and give the expected answer.
    """
    A = Parameter('A')
    x = Variable('x')
    new = A * x ** 2

    assert new(x=2, A=2) == 8
    assert not new(x=2, A=3) == 8


def test_simple_sigma():
    """
    Make sure we produce the same results as scipy's curve_fit, with and
    without sigmas, and compare the results of both to a known value.
    """
    t_data = np.array([1.4, 2.1, 2.6, 3.0, 3.3])
    y_data = np.array([10, 20, 30, 40, 50])

    sigma = 0.2
    n = np.array([5, 3, 8, 15, 30])
    sigma_t = sigma / np.sqrt(n)

    # We now define our model
    y = Variable('x')
    g = Parameter('g')
    t_model = (2 * y / g)**0.5

    fit = Fit(t_model, y_data, t_data)  # , sigma=sigma_t)
    fit_result = fit.execute()

    # h_smooth = np.linspace(0,60,100)
    # t_smooth = t_model(y=h_smooth, **fit_result.params)

    # Lets with the results from curve_fit, no weights
    popt_noweights, pcov_noweights = curve_fit(lambda y, p: (2 * y / p)**0.5, y_data, t_data)

    assert fit_result.value(g) == pytest.approx(popt_noweights[0])
    assert fit_result.stdev(g) == pytest.approx(np.sqrt(pcov_noweights[0, 0]), 1e-6)

    # Same sigma everywere
    fit = Fit(t_model, y_data, t_data, 0.0031, absolute_sigma=False)
    fit_result = fit.execute()
    popt_sameweights, pcov_sameweights = curve_fit(
        lambda y, p: (2 * y / p)**0.5,
        y_data,
        t_data,
        sigma=0.0031*np.ones(len(y_data)),
        absolute_sigma=False
    )
    assert fit_result.value(g) == pytest.approx(popt_sameweights[0], 1e-4)
    assert fit_result.stdev(g) == pytest.approx(np.sqrt(pcov_sameweights[0, 0]), 1e-4)
    # Same weight everywere should be the same as no weight when absolute_sigma=False
    assert fit_result.value(g) == pytest.approx(popt_noweights[0], 1e-4)
    assert fit_result.stdev(g) == pytest.approx(np.sqrt(pcov_noweights[0, 0]), 1e-4)

    # Different sigma for every point
    fit = Fit(t_model, y_data, t_data, 0.1*sigma_t, absolute_sigma=False)
    fit_result = fit.execute()
    popt, pcov = curve_fit(lambda y, p: (2 * y / p)**0.5, y_data, t_data, sigma=.1*sigma_t)

    assert fit_result.value(g) == pytest.approx(popt[0])
    assert fit_result.stdev(g) == pytest.approx(np.sqrt(pcov[0, 0]), 1e-6)

    # according to Mathematica
    assert fit_result.value(g) == pytest.approx(9.095, 1e-3)
    assert fit_result.stdev(g) == pytest.approx(0.102, 1e-2)


def test_error_advanced():
    """
    Models an example from the mathematica docs and try's to replicate it
    using both symfit and scipy's curve_fit.
    http://reference.wolfram.com/language/howto/FitModelsWithMeasurementErrors.html
    """
    data = [
        [0.9, 6.1, 9.5], [3.9, 6., 9.7], [0.3, 2.8, 6.6],
        [1., 2.2, 5.9], [1.8, 2.4, 7.2], [9., 1.7, 7.],
        [7.9, 8., 10.4], [4.9, 3.9, 9.], [2.3, 2.6, 7.4],
        [4.7, 8.4, 10.]
    ]
    xdata, ydata, zdata = [np.array(data) for data in zip(*data)]
    xy = np.vstack((xdata, ydata))
    errors = np.array([.4, .4, .2, .4, .1, .3, .1, .2, .2, .2])

    # raise Exception(xy, z)
    a = Parameter('a', value=3.0)
    b = Parameter('b', value=0.9)
    c = Parameter('c', value=5)
    x = Variable('x')
    y = Variable('y')
    z = Variable('z')
    model = {z: a * log(b * x + c * y)}

    # Use a gradient model because Mathematica uses the Hessian
    # approximation instead of the exact Hessian.
    model = GradientModel(model)
    fit = Fit(model, x=xdata, y=ydata, z=zdata, absolute_sigma=False)
    fit_result = fit.execute()

    # Same as Mathematica default behavior.
    assert fit_result.value(a) == pytest.approx(2.9956, 1e-4)
    assert fit_result.value(b) == pytest.approx(0.563212, 1e-4)
    assert fit_result.value(c) == pytest.approx(3.59732, 1e-4)
    assert fit_result.stdev(a) == pytest.approx(0.278304, 1e-4)
    assert fit_result.stdev(b) == pytest.approx(0.224107, 1e-4)
    assert fit_result.stdev(c) == pytest.approx(0.980352, 1e-4)

    fit = Fit(model, xdata, ydata, zdata, absolute_sigma=True)
    fit_result = fit.execute()
    # Same as Mathematica in Measurement error mode, but without suplying
    # any errors.
    assert fit_result.value(a) == pytest.approx(2.9956, 1e-4)
    assert fit_result.value(b) == pytest.approx(0.563212, 1e-4)
    assert fit_result.value(c) == pytest.approx(3.59732, 1e-4)
    assert fit_result.stdev(a) == pytest.approx(0.643259, 1e-4)
    assert fit_result.stdev(b) == pytest.approx(0.517992, 1e-4)
    assert fit_result.stdev(c) == pytest.approx(2.26594, 1e-4)

    fit = Fit(model, xdata, ydata, zdata, sigma_z=errors)
    fit_result = fit.execute()

    popt, pcov, infodict, errmsg, ier = curve_fit(
        lambda x_vec, a, b, c: a * np.log(b * x_vec[0] + c * x_vec[1]),
        xy, zdata, sigma=errors, absolute_sigma=True, full_output=True
    )

    # Same as curve_fit?
    assert fit_result.value(a) == pytest.approx(popt[0], 1e-4)
    assert fit_result.value(b) == pytest.approx(popt[1], 1e-4)
    assert fit_result.value(c) == pytest.approx(popt[2], 1e-4)
    assert fit_result.stdev(a) == pytest.approx(np.sqrt(pcov[0, 0]), 1e-4)
    assert fit_result.stdev(b) == pytest.approx(np.sqrt(pcov[1, 1]), 1e-4)
    assert fit_result.stdev(c) == pytest.approx(np.sqrt(pcov[2, 2]), 1e-4)

    # Same as Mathematica with MEASUREMENT ERROR
    assert fit_result.value(a) == pytest.approx(2.68807, 1e-4)
    assert fit_result.value(b) == pytest.approx(0.941344, 1e-4)
    assert fit_result.value(c) == pytest.approx(5.01541, 1e-4)
    assert fit_result.stdev(a) == pytest.approx(0.0974628, 1e-4)
    assert fit_result.stdev(b) == pytest.approx(0.247018, 1e-4)
    assert fit_result.stdev(c) == pytest.approx(0.597661, 1e-4)


def test_error_analytical():
    """
    Test using a case where the analytical answer is known. Uses both
    symfit and scipy's curve_fit.
    Modeled after:
    http://nbviewer.ipython.org/urls/gist.github.com/taldcroft/5014170/raw/31e29e235407e4913dc0ec403af7ed524372b612/curve_fit.ipynb
    """
    N = 10000
    sigma = 10.0 * np.ones(N)
    xn = np.arange(N, dtype=float)
    # yn = np.zeros_like(xn)
    np.random.seed(10)
    yn = np.random.normal(size=len(xn), scale=sigma)

    a = Parameter('a')
    y = Variable('y')
    model = {y: a}

    fit = Fit(model, y=yn, sigma_y=sigma)
    fit_result = fit.execute()

    popt, pcov = curve_fit(lambda x, a: a * np.ones_like(x),
                           xn, yn, sigma=sigma, absolute_sigma=True)
    assert fit_result.value(a) == pytest.approx(popt[0], 1e-5)
    assert fit_result.stdev(a) == pytest.approx(np.sqrt(np.diag(pcov))[0], 1e-2)

    fit_no_sigma = Fit(model, yn)
    fit_result_no_sigma = fit_no_sigma.execute()

    popt, pcov = curve_fit(lambda x, a: a * np.ones_like(x), xn, yn,)
    # With or without sigma, the bestfit params should be in agreement in case of equal weights
    assert fit_result.value(a) == pytest.approx(fit_result_no_sigma.value(a), 1e-5)
    # Since symfit is all about absolute errors, the sigma will not be in agreement
    assert not fit_result.stdev(a) == fit_result_no_sigma.stdev(a) == 5
    assert fit_result_no_sigma.stdev(a) == pytest.approx(pcov[0][0]**0.5, 1e-5)
    assert fit_result_no_sigma.value(a) == pytest.approx(popt[0], 1e-5)

    # Analytical answer for mean of N(0,1):
    mu = 0.0
    sigma_mu = sigma[0]/N**0.5

    assert fit_result.stdev(a) == pytest.approx(sigma_mu, 1e-5)

# TODO: redudant with test_error_analytical?
@pytest.mark.skip()
def test_straight_line_analytical():
     """
     Test symfit against a straight line, for which the parameters and their
     uncertainties are known analytically. Assuming equal weights.
     """
     data = [[0, 1], [1, 0], [3, 2], [5, 4]]
     x, y = (np.array(i, dtype='float64') for i in zip(*data))
     # x = np.arange(0, 100, 0.1)
     # np.random.seed(10)
     # y = 3.0*x + 105.0 + np.random.normal(size=x.shape)

     dx = x - x.mean()
     dy = y - y.mean()
     mean_squared_x = np.mean(x**2) - np.mean(x)**2
     mean_xy = np.mean(x * y) - np.mean(x)*np.mean(y)
     a = mean_xy/mean_squared_x
     b = y.mean() - a * x.mean()
     assert a == pytest.approx(0.694915, 1e-6) # values form Mathematica
     assert b == pytest.approx(0.186441, 1e-6)

     S = np.sum((y - (a*x + b))**2)
     var_a_exact = S/(len(x) * (len(x) - 2) * mean_squared_x)
     var_b_exact = var_a_exact*np.mean(x ** 2)
     a_exact = a
     b_exact = b

     # We will now compare these exact results with values from symfit
     a, b, x_var = Parameter(name='a', value=3.0), Parameter(name='b'), Variable(name='x')
     model = a*x_var + b
     fit = Fit(model, x, y, absolute_sigma=False)
     fit_result = fit.execute()

     popt, pcov = curve_fit(lambda z, c, d: c * z + d, x, y,
                            Dfun=lambda p, x, y, func: np.transpose([x, np.ones_like(x)]))
                             # Dfun=lambda p, x, y, func: print(p, func, x, y))

     # curve_fit
     assert a_exact == pytest.approx(popt[0], 1e-4)
     assert b_exact == pytest,approx(popt[1], 1e-4)
     assert var_a_exact == pytest.approx(pcov[0][0], 1e-6)
     assert var_b_exact == pytest.approx(pcov[1][1], 1e-6)

     assert a_exact == pytest.approx(fit_result.params.a, 1e-4)
     assert b_exact == pytest.approx(fit_result.params.b, 1e-4)
     assert var_a_exact**0.5 == pytest.approx(fit_result.params.a_stdev, 1e-6)
     assert var_b_exact**0.5 == pytest.approx(fit_result.params.b_stdev, 1e-6)


def test_fixed_parameters():
    """
    Make sure fixed parameters don't change on fitting
    """
    a, b, c, d = parameters('a, b, c, d')
    x, y = variables('x, y')

    c.value = 4.0
    a.min, a.max = 1.0, 5.0  # Bounds are needed for DifferentialEvolution
    b.min, b.max = 1.0, 5.0
    c.min, c.max = 1.0, 5.0
    d.min, d.max = 1.0, 5.0
    c.fixed = True

    model = Model({y: a * exp(-(x - b)**2 / (2 * c**2)) + d})
    # Generate data
    xdata = np.linspace(0, 100)
    ydata = model(xdata, a=2, b=3, c=2, d=2).y

    for minimizer in subclasses(BaseMinimizer):
        if minimizer is ChainedMinimizer:
            continue
        else:
            fit = Fit(model, x=xdata, y=ydata, minimizer=minimizer)
            fit_result = fit.execute()
            # Should still be 4.0, not 2.0!
            assert 4.0 == fit_result.params['c']


def test_fixed_parameters_2():
    """
    Make sure parameter boundaries are respected
    """
    x = Parameter('x', min=1)
    y = Variable('y')
    model = Model({y: x**2})

    bounded_minimizers = list(subclasses(BoundedMinimizer))
    for minimizer in bounded_minimizers:
        if minimizer is MINPACK:
            # Not a MINPACKable problem because it only has a param
            continue
        fit = Fit(model, minimizer=minimizer)
        assert isinstance(fit.objective, MinimizeModel)
        if minimizer is DifferentialEvolution:
            # Also needs a max
            x.max = 10
            fit_result = fit.execute()
            x.max = None
        else:
            fit_result = fit.execute()
            assert fit_result.value(x) >= 1.0
            assert fit_result.value(x) <= 2.0
        assert fit.minimizer.bounds == [(1, None)]


def test_non_boundaries():
    """
    Make sure parameter boundaries are not invented
    """
    x = Parameter('x')
    y = Variable('y')
    model = Model({y: x**2})

    bounded_minimizers = list(subclasses(BoundedMinimizer))
    bounded_minimizers = [minimizer for minimizer in bounded_minimizers
                          if minimizer is not DifferentialEvolution]
    for minimizer in bounded_minimizers:
        # Not a MINPACKable problem because it only has a param
        if minimizer is MINPACK:
            continue
        fit = Fit(model, minimizer=minimizer)
        fit_result = fit.execute()
        assert fit_result.value(x) == pytest.approx(0.0)
        assert fit.minimizer.bounds == [(None, None)]


def test_single_param_model():
    """
    Added after #161, this tests if models with a single additive parameter
    are fitted properly. The problem with these models is that their
    jacobian is in principle just int 1, which is not the correct shape.

    No news is good news.
    :return:
    """
    T = Variable('T')
    l = Variable('l')
    s = Parameter('s', value=300)
    a = Parameter('a', value=300)
    model = {l: s + a + 1 / (1 + exp(- T))}

    temp_data = [270, 280, 285, 290, 295, 300, 310, 320]
    length_data = [8.33, 8.41, 8.45, 8.5, 8.54, 9.13, 9.27, 9.4]
    fit = Fit(model, l=length_data, T=temp_data)
    fit_result = fit.execute()

    # Raise the stakes by increasing the dimensionality of the data
    TT, LL = np.meshgrid(temp_data, length_data)
    fit = Fit(model, l=LL, T=TT)
    fit_result = fit.execute()


def test_model_from_dict():
    """
    Tries to create a model from a dictionary.
    """
    x, y_1, y_2 = variables('x, y_1, y_2')
    a, b = parameters('a, b')
    # This way the test fails rather than errors.
    try:
        Model({
            y_1: 2 * a * x,
            y_2: b * x**2
        })
    except Exception as error:
        pytest.fail('test_model_from_dict raised {}'.format(error))


def test_version():
    """
    Test if __version__ is availabe
    :return:
    """
    import symfit
    symfit.__version__



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

from __future__ import division, print_function
import pytest
import pickle
from collections import OrderedDict

import numpy as np
from symfit import (
    Variable, Parameter, Fit, FitResults, Eq, Ge, CallableNumericalModel, Model
)
from symfit.distributions import BivariateGaussian
from symfit.core.minimizers import (
    BaseMinimizer, MINPACK, BFGS, NelderMead, ChainedMinimizer, BasinHopping
)
from symfit.core.objectives import (
    LogLikelihood, LeastSquares, VectorLeastSquares, MinimizeModel
)


def ge_constraint(a):  # Has to be in the global namespace for pickle.
    return a - 1


class TestTestResult():

    @classmethod
    def setup_class(cls):
        xdata = np.linspace(1, 10, 10)
        ydata = 3 * xdata ** 2

        cls.a = Parameter('a')
        cls.b = Parameter('b')

        x = Variable('x')
        y = Variable('y')
        model = Model({y: cls.a * x ** cls.b})

        fit = Fit(model, x=xdata, y=ydata)
        cls.fit_result = fit.execute()
        fit = Fit(model, x=xdata, y=ydata, minimizer=MINPACK)
        cls.minpack_result = fit.execute()
        fit = Fit(model, x=xdata, objective=LogLikelihood)
        cls.likelihood_result = fit.execute()
        fit = Fit(model, x=xdata, y=ydata, minimizer=[BFGS, NelderMead])
        cls.chained_result = fit.execute()

        z = Variable('z')
        constraints = [
            Eq(cls.a, cls.b),
            CallableNumericalModel.as_constraint(
                {z: ge_constraint}, connectivity_mapping={z: {cls.a}},
                constraint_type=Ge, model=model
            )
        ]
        fit = Fit(model, x=xdata, y=ydata, constraints=constraints)
        cls.constrained_result = fit.execute()
        fit = Fit(model, x=xdata, y=ydata, constraints=constraints,
                  minimizer=BasinHopping)
        cls.constrained_basinhopping_result = fit.execute()

    def test_params_type(self):
        assert isinstance(self.fit_result.params, OrderedDict)

    def test_minimizer_output_type(self):
        assert isinstance(self.fit_result.minimizer_output, dict)
        assert isinstance(self.minpack_result.minimizer_output, dict)
        assert isinstance(self.likelihood_result.minimizer_output, dict)

    def test_fitting(self):
        """
        Test if the fitting worked in the first place.
        """
        assert isinstance(self.fit_result, FitResults)
        assert self.fit_result.value(self.a) == pytest.approx(3.0)
        assert self.fit_result.value(self.b) == pytest.approx(2.0)

        assert isinstance(self.fit_result.stdev(self.a), float)
        assert isinstance(self.fit_result.stdev(self.b), float)

        assert isinstance(self.fit_result.r_squared, float)
        # by definition since there's no fuzzyness
        assert self.fit_result.r_squared == 1.0

    def test_fitting_2(self):
        np.random.seed(43)
        mean = (0.62, 0.71)  # x, y mean 0.7, 0.7
        cov = [
            [0.102**2, 0],
            [0, 0.07**2]
        ]
        data_1 = np.random.multivariate_normal(mean, cov, 10**5)
        mean = (0.33, 0.28)  # x, y mean 0.3, 0.3
        cov = [  # rho = 0.25
            [0.05 ** 2, 0.25 * 0.05 * 0.101],
            [0.25 * 0.05 * 0.101, 0.101 ** 2]
        ]
        data_2 = np.random.multivariate_normal(mean, cov, 10**5)
        data = np.vstack((data_1, data_2))

        # Insert them as y,x here as np fucks up cartesian conventions.
        ydata, xedges, yedges = np.histogram2d(data[:, 1], data[:, 0], bins=200,
                                               range=[[0.0, 1.0], [0.0, 1.0]],
                                               density=True)
        xcentres = (xedges[:-1] + xedges[1:]) / 2
        ycentres = (yedges[:-1] + yedges[1:]) / 2

        # Make a valid grid to match ydata
        xx, yy = np.meshgrid(xcentres, ycentres, sparse=False)

        x = Variable('x')
        y = Variable('y')

        x0_1 = Parameter('x0_1', value=0.6, min=0.5, max=0.7)
        sig_x_1 = Parameter('sig_x_1', value=0.1, min=0.0, max=0.2)
        y0_1 = Parameter('y0_1', value=0.7, min=0.6, max=0.8)
        sig_y_1 = Parameter('sig_y_1', value=0.05, min=0.0, max=0.2)
        rho_1 = Parameter('rho_1', value=0.0, min=-0.5, max=0.5)
        A_1 = Parameter('A_1', value=0.5, min=0.3, max=0.7)
        g_1 = A_1 * BivariateGaussian(x=x, y=y, mu_x=x0_1, mu_y=y0_1,
                                      sig_x=sig_x_1, sig_y=sig_y_1, rho=rho_1)

        x0_2 = Parameter('x0_2', value=0.3, min=0.2, max=0.4)
        sig_x_2 = Parameter('sig_x_2', value=0.05, min=0.0, max=0.2)
        y0_2 = Parameter('y0_2', value=0.3, min=0.2, max=0.4)
        sig_y_2 = Parameter('sig_y_2', value=0.1, min=0.0, max=0.2)
        rho_2 = Parameter('rho_2', value=0.26, min=0.0, max=0.8)
        A_2 = Parameter('A_2', value=0.5, min=0.3, max=0.7)
        g_2 = A_2 * BivariateGaussian(x=x, y=y, mu_x=x0_2, mu_y=y0_2,
                                      sig_x=sig_x_2, sig_y=sig_y_2, rho=rho_2)

        model = g_1 + g_2
        fit = Fit(model, xx, yy, ydata)
        fit_result = fit.execute()

        assert fit_result.r_squared > 0.95
        for param in fit.model.params:
            try:
                assert fit_result.stdev(param)**2 == pytest.approx(fit_result.variance(param))
            except AssertionError:
                assert fit_result.variance(param) <= 0.0
                assert np.isnan(fit_result.stdev(param))

        # Covariance matrix should be symmetric
        for param_1 in fit.model.params:
            for param_2 in fit.model.params:
                assert fit_result.covariance(param_1, param_2) == pytest.approx(fit_result.covariance(param_2, param_1), rel=1e-3)

    def test_minimizer_included(self):
        """"The minimizer used should be included in the results."""
        assert isinstance(self.constrained_result.minimizer, BaseMinimizer)
        assert isinstance(self.constrained_basinhopping_result.minimizer, BaseMinimizer)
        assert isinstance(self.likelihood_result.minimizer, BaseMinimizer)
        assert isinstance(self.fit_result.minimizer, BaseMinimizer)
        assert isinstance(self.chained_result.minimizer, ChainedMinimizer)
        for minimizer, cls in zip(self.chained_result.minimizer.minimizers, [BFGS, NelderMead]):
            assert isinstance(minimizer, cls)

    def test_objective_included(self):
        """"The objective used should be included in the results."""
        assert isinstance(self.fit_result.objective, LeastSquares)
        assert isinstance(self.minpack_result.objective, VectorLeastSquares)
        assert isinstance(self.likelihood_result.objective, LogLikelihood)
        assert isinstance(self.constrained_result.objective, LeastSquares)
        assert isinstance(self.constrained_basinhopping_result.objective, LeastSquares)

    def test_constraints_included(self):
        """
        Test if the constraints have been properly fed to the results object so
        we can easily print their compliance.
        """
        # For a constrained fit we expect a list of MinimizeModel objectives.
        for constrained_result in [self.constrained_result, self.constrained_basinhopping_result]:
            assert isinstance(constrained_result.constraints, list)
            for constraint in self.constrained_result.constraints:
                assert isinstance(constraint, MinimizeModel)

    def test_message_included(self):
        """Status message should be included."""
        assert isinstance(self.fit_result.status_message, str)
        assert isinstance(self.minpack_result.status_message, str)
        assert isinstance(self.likelihood_result.status_message, str)
        assert isinstance(self.constrained_result.status_message, str)
        assert isinstance(self.constrained_basinhopping_result.status_message, str)

    def test_pickle(self):
        for fit_result in [self.fit_result, self.chained_result,
                           self.constrained_basinhopping_result,
                           self.constrained_result, self.likelihood_result]:
            dumped = pickle.dumps(fit_result)
            new_result = pickle.loads(dumped)
            assert sorted(fit_result.__dict__.keys()) == sorted(new_result.__dict__.keys())
            for k, v1 in fit_result.__dict__.items():
                v2 = new_result.__dict__[k]
                if k == 'minimizer':
                    assert type(v1) == type(v2)
                elif k != 'minimizer_output':  # Ignore minimizer_output
                    if isinstance(v1, np.ndarray):
                        assert v1 == pytest.approx(v2, nan_ok=True)

    def test_gof_presence(self):
        """
        Test if the expected goodness of fit estimators are present.
        """
        assert hasattr(self.fit_result, 'objective_value')
        assert hasattr(self.fit_result, 'r_squared')
        assert hasattr(self.fit_result, 'chi_squared')
        assert not hasattr(self.fit_result, 'log_likelihood')
        assert not hasattr(self.fit_result, 'likelihood')

        assert hasattr(self.minpack_result, 'objective_value')
        assert hasattr(self.minpack_result, 'r_squared')
        assert hasattr(self.minpack_result, 'chi_squared')
        assert not hasattr(self.minpack_result, 'log_likelihood')
        assert not hasattr(self.minpack_result, 'likelihood')

        assert hasattr(self.likelihood_result, 'objective_value')
        assert not hasattr(self.likelihood_result, 'r_squared')
        assert not hasattr(self.likelihood_result, 'chi_squared')
        assert hasattr(self.likelihood_result, 'log_likelihood')
        assert hasattr(self.likelihood_result, 'likelihood')



# SPDX-FileCopyrightText: 2014-2020 Martin Roelfs
#
# SPDX-License-Identifier: MIT

# -*- coding: utf-8 -*-

from symfit import variables, Parameter, exp, Fit, Model
from symfit.distributions import Gaussian
from symfit.contrib.interactive_guess import InteractiveGuess
import numpy as np


x, y, z = variables('x, y, z')
mu_x = Parameter('mu_x', 10)
mu_y = Parameter('mu_y', 10)
sig_x = Parameter('sig_x', 1)
sig_y = Parameter('sig_y', 1)



model = Model({z: Gaussian(x, mu_x, sig_x) * Gaussian(y, mu_y, sig_y)})
x_data = np.linspace(0, 25, 50)
y_data = np.linspace(0, 25, 50)
x_data, y_data = np.meshgrid(x_data, y_data)
x_data = x_data.flatten()
y_data = y_data.flatten()
z_data = model(x=x_data, y=y_data, mu_x=5, sig_x=0.3, mu_y=10, sig_y=1).z

guess = InteractiveGuess(model, x=x_data, y=y_data, z=z_data)
guess.execute()
print(guess)

fit = Fit(model, x=x_data, y=y_data, z=z_data)
fit_result = fit.execute()
print(fit_result)
